###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
##TODO Tests
""" This module defines the class Pipeline """
from __future__ import division

from vistrails.core.cache.hasher import Hasher
from vistrails.core.configuration import get_vistrails_configuration
from vistrails.core.data_structures.bijectivedict import Bidict
from vistrails.core.data_structures.graph import Graph
from vistrails.core import debug
from vistrails.core.modules.module_registry import get_module_registry, \
    ModuleRegistryException, MissingModuleVersion, MissingPackage, PortMismatch
from vistrails.core.system import get_vistrails_basic_pkg_id
from vistrails.core.utils import VistrailsInternalError
from vistrails.core.vistrail.group import Group
from vistrails.core.vistrail.module_control_param import ModuleControlParam
from vistrails.core.vistrail.plugin_data import PluginData
from vistrails.core.vistrail.port_spec import PortSpec
from vistrails.db.domain import DBWorkflow
import vistrails.core.vistrail.action
from vistrails.core.utils import InvalidPipeline

import copy

import unittest
from vistrails.core.vistrail.abstraction import Abstraction
from vistrails.core.vistrail.connection import Connection
from vistrails.core.vistrail.location import Location
from vistrails.core.vistrail.module import Module
from vistrails.core.vistrail.module_function import ModuleFunction
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.core.vistrail.port import Port
from vistrails.db.domain import IdScope

##############################################################################

class MissingVistrailVariable(Exception):
    def __init__(self, var_uuid, identifier, name, namespace):
        self._var_uuid = var_uuid
        self._identifier = identifier
        self._name = name
        self._namespace = namespace

    def __str__(self):
        return "Missing Vistrail Variable '%s' of type %s from package %s" % (self._var_uuid,
                self._module_name, self._identifier)
        
    def __eq__(self, other):
        return type(self) == type(other) and \
            self._var_uuid == other._var_uuid and \
            self._identifier == other._identifier and \
            self._name == other._name and \
            self._namespace == other._namespace

    def __hash__(self):
        return (type(self), self._var_uuid, self._identifier,
                self._name, self._namespace).__hash__()

    def _get_module_name(self):
        if self._namespace:
            return "%s|%s" % (self._namespace, self._name)
        return self._name
    _module_name = property(_get_module_name)

class MissingFunction(Exception):
    def __init__(self, name, module_name, module_id=None):
        self.name = name
        self.module_name = module_name
        self.module_id = module_id

    def __str__(self):
        return ("Missing Function '%s' on module '%s'%s" % 
                (self.name, self.module_name, " (id %d)" % self.module_id if
                 self.module_id is not None else ""))

class CycleInPipeline(Exception):
    def __str__(self):
        return "Pipeline contains a cycle"

class Pipeline(DBWorkflow):
    """ A Pipeline is a set of modules and connections between them. """
    
    def __init__(self, *args, **kwargs):
        """ __init__() -> Pipelines
        Initializes modules, connections and graph.

        """
        self.clear()

        DBWorkflow.__init__(self, *args, **kwargs)
        if self.id is None:
            self.id = 0
        if self.name is None:
            self.name = 'untitled'
        self.set_defaults()

    def set_defaults(self, other=None):
        if other is None:
            self.is_valid = False
            self.aliases = Bidict()
            self._subpipeline_signatures = Bidict()
            self._module_signatures = Bidict()
            self._connection_signatures = Bidict()
        else:
            self.is_valid = other.is_valid
            self.aliases = Bidict([(k,copy.copy(v))
                                   for (k,v) in other.aliases.iteritems()])
            self._connection_signatures = \
                Bidict([(k,copy.copy(v))
                        for (k,v) in other._connection_signatures.iteritems()])
            self._subpipeline_signatures = \
                Bidict([(k,copy.copy(v))
                        for (k,v) in other._subpipeline_signatures.iteritems()])
            self._module_signatures = \
                Bidict([(k,copy.copy(v))
                        for (k,v) in other._module_signatures.iteritems()])

        self.graph = Graph()
        for module in self.module_list:
            self.graph.add_vertex(module.id)
            # there should be another way to do this
            m_id = module.id
            for fun in module.functions:
                for par in fun.parameters:
                    self.change_alias(par.alias,
                                      par.vtType,
                                      par.real_id,
                                      fun.vtType,
                                      fun.real_id,
                                      m_id)
            module.connected_input_ports = {}
            module.connected_output_ports = {}

        for connection in self.connection_list:
            self.graph.add_edge(connection.source.moduleId,
                                connection.destination.moduleId,
                                connection.id)
            c = connection
            source_name = c.source.name
            output_ports = self.modules[c.sourceId].connected_output_ports
            if source_name not in output_ports:
                output_ports[source_name] = 0
            output_ports[source_name] += 1
                
            dest_name = c.destination.name
            input_ports = self.modules[c.destinationId].connected_input_ports
            if dest_name not in input_ports:
                input_ports[dest_name] = 0
            input_ports[dest_name] += 1
            
    def __copy__(self):
        """ __copy__() -> Pipeline - Returns a clone of itself """ 
        return Pipeline.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBWorkflow.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = Pipeline
        cp.set_defaults(self)
        return cp

    @staticmethod
    def convert(_workflow):
        if _workflow.__class__ == Pipeline:
            return
        # do clear plus get the modules and connections
        _workflow.__class__ = Pipeline
        for _module in _workflow.db_modules:
            if _module.vtType == Module.vtType:
                Module.convert(_module)
            elif _module.vtType == Abstraction.vtType:
                Abstraction.convert(_module)
            elif _module.vtType == Group.vtType:
                Group.convert(_module)
        for _connection in _workflow.db_connections:
            Connection.convert(_connection)
        for _plugin_data in _workflow.db_plugin_datas:
            PluginData.convert(_plugin_data)
        _workflow.set_defaults()

    ##########################################################################

    def find_method(self, module_id, parameter_name):
        """find_method(module_id, parameter_name) -> int.

        Finds the function_id for a given method name.
        Returns -1 if method name is not there.

        WARNING: Might not work for overloaded methods (where types
        also matter)
        """
        try:
            return [f.name
                    for f
                    in self.get_module_by_id(module_id).functions].index(parameter_name)
        except ValueError:
            return -1

    ##########################################################################
    # Properties

    id = DBWorkflow.db_id
    name = DBWorkflow.db_name
    plugin_datas = DBWorkflow.db_plugin_datas

    def _get_modules(self):
        return self.db_modules_id_index
    modules = property(_get_modules)
    def _get_module_list(self):
        return self.db_modules
    module_list = property(_get_module_list)

    def _get_connections(self):
        return self.db_connections_id_index
    connections = property(_get_connections)
    def _get_connection_list(self):
        return self.db_connections
    connection_list = property(_get_connection_list)

    def clear(self):
        """clear() -> None. Erases pipeline contents."""
        if hasattr(self, 'db_modules'):
            for module in self.db_modules:
                self.db_delete_module(module)
        if hasattr(self, 'db_connections'):
            for connection in self.db_connections:
                self.db_delete_connection(connection)
        self.graph = Graph()
        self.aliases = Bidict()
        self._subpipeline_signatures = Bidict()
        self._module_signatures = Bidict()
        self._connection_signatures = Bidict()

    def get_tmp_id(self, type):
        """get_tmp_id(type: str) -> long
        returns a temporary id for a workflow item.  Use the idScope on the
        vistrail for permanent ids.
        """

        return -self.tmp_id.getNewId(type)

    def fresh_module_id(self):
        return self.get_tmp_id(Module.vtType)
    def fresh_connection_id(self):
        return self.get_tmp_id(Connection.vtType)

    def connections_to_module(self, moduleId):
        """ connections_to_module(int moduleId) -> list of module ids
        returns a list of module ids that are inputs to the given moduleId

        """

        modules = []
        for edge in self.graph.edges_to(moduleId):
            modules.append(self.modules[edge[0]].id)
        return modules

    def get_inputPort_modules(self, moduleId, portName):
        """ get_inputPort_modules(int moduleId, string portName)-> list of module ids
        returns a list of module ids that are the input to a given port
        on a given module

        """
        modules = []
        for edge in self.graph.edges_to(moduleId):
            if self.connections[edge[1]].destination.name == portName:
                modules.append(self.modules[edge[0]].id)
        return modules

    def get_outputPort_modules(self, moduleId, portName):
        """ get_outputPort_modules(int moduleId, string portName)-> list of module ids
        returns a list of module ids that are the output to a given port
        on a given module
        """
        modules = []
        for edge in self.graph.edges_from(moduleId):
            if self.connections[edge[1]].source.name == portName:
                modules.append(self.modules[edge[0]].id)
        return modules

    def perform_action_chain(self, actionChain):
        # BEWARE: if actionChain is long, you're probably better off
        # going through general_action_chain, because it optimizes
        # away unnecessary operations.
        for action in actionChain:
            self.perform_action(action)

    def perform_action(self, action):
#         print "+++"
#         for operation in action.operations:
#             print operation.vtType, operation.what, operation.old_obj_id, \
#                 operation.new_obj_id, operation.parentObjType, operation.parentObjId
#         print "---"
        for operation in action.operations:
            self.perform_operation(operation)

    def perform_operation_chain(self, opChain):
        for op in opChain:
            self.perform_operation(op)

    def perform_operation(self, op):
        # print "doing %s %s %s" % (op.id, op.vtType, op.what)
        if op.db_what == 'abstraction' or op.db_what == 'group':
            what = 'module'
        else:
            what = op.db_what
        funname = '%s_%s' % (op.vtType, what)
        try:
            f = getattr(self, funname)
        except AttributeError:
            db_funname = 'db_%s_object' % op.vtType
            try:
                f = getattr(self, db_funname)
            except AttributeError:
                msg = "Pipeline cannot execute '%s %s' operation" % \
                    (op.vtType, op.what)
                raise VistrailsInternalError(msg)

        if op.vtType == 'add':
            f(op.data, op.parentObjType, op.parentObjId)
        elif op.vtType == 'delete':
            f(op.objectId, op.what, op.parentObjType, op.parentObjId)
        elif op.vtType == 'change':
            f(op.oldObjId, op.data, op.parentObjType, op.parentObjId)

    def add_module(self, m, *args):
        """add_module(m: Module) -> None 
        Add new module to pipeline
          
        """
        if self.has_module_with_id(m.id):
            raise VistrailsInternalError("duplicate module id: %d" % m.id )
#         self.modules[m.id] = copy.copy(m)
#         if m.vtType == Abstraction.vtType:
#             m.abstraction = self.abstraction_map[m.abstraction_id]
        self.db_add_object(m)
        self.graph.add_vertex(m.id)

    def change_module(self, old_id, m, *args):
        if not self.has_module_with_id(old_id):
            raise VistrailsInternalError("module %s doesn't exist" % old_id)
        self.db_change_object(old_id, m)
        self.graph.delete_vertex(old_id)
        self.graph.add_vertex(m.id)

    def delete_module(self, id, *args):
        """delete_module(id:int) -> None 
        Delete a module from pipeline given an id.

        """
        if not self.has_module_with_id(id):
            raise VistrailsInternalError("id missing in modules")

        # we're hiding the necessary operations by doing this!
        for (_, conn_id) in self.graph.adjacency_list[id][:]:
            self.delete_connection(conn_id)
        for (_, conn_id) in self.graph.inverse_adjacency_list[id][:]:
            self.delete_connection(conn_id)

        # self.modules.pop(id)
        self.db_delete_object(id, Module.vtType)
        self.graph.delete_vertex(id)
        if id in self._module_signatures:
            del self._module_signatures[id]
        if id in self._subpipeline_signatures:
            del self._subpipeline_signatures[id]

    def add_connection(self, c, *args):
        """add_connection(c: Connection) -> None 
        Add new connection to pipeline.
          
        """
        if self.has_connection_with_id(c.id):
            raise VistrailsInternalError("duplicate connection id " + str(c.id))
#         self.connections[c.id] = copy.copy(c)
        self.db_add_object(c)
        if c.source is not None and c.destination is not None:
            assert(c.sourceId != c.destinationId)        
            self.graph.add_edge(c.sourceId, c.destinationId, c.id)
            self.ensure_connection_specs([c.id])

            source_name = c.source.name
            output_ports = self.modules[c.sourceId].connected_output_ports
            if source_name not in output_ports:
                output_ports[source_name] = 0
            output_ports[source_name] += 1
                
            dest_name = c.destination.name
            input_ports = self.modules[c.destinationId].connected_input_ports
            if dest_name not in input_ports:
                input_ports[dest_name] = 0
            input_ports[dest_name] += 1

    def change_connection(self, old_id, c, *args):
        """change_connection(old_id: long, c: Connection) -> None
        Deletes connection identified by old_id and adds connection c

        """
        if not self.has_connection_with_id(old_id):
            raise VistrailsInternalError("connection %s doesn't exist" % old_id)

        old_conn = self.connections[old_id]
        if old_conn.source is not None and old_conn.destination is not None:
            self.graph.delete_edge(old_conn.sourceId, old_conn.destinationId,
                                   old_conn.id)
            if self.graph.out_degree(old_conn.sourceId) < 1:
                self.modules[old_conn.sourceId].connected_output_ports.discard(
                    old_conn.source.name)
            if self.graph.in_degree(old_conn.destinationId) < 1:
                connected_input_ports = \
                    self.modules[old_conn.destinationId].connected_input_ports
                connected_input_ports.discard(conn.destination.name)

        if old_id in self._connection_signatures:
            del self._connection_signatures[old_id]
        self.db_change_object(old_id, c)        
        if c.source is not None and c.destination is not None:
            assert(c.sourceId != c.destinationId)
            self.graph.add_edge(c.sourceId, c.destinationId, c.id)
            self.ensure_connection_specs([c.id])
            self.modules[c.sourceId].connected_output_ports.add(c.source.name)
            self.modules[c.destinationId].connected_input_ports.add(
                c.destination.name)

    def delete_connection(self, id, *args):
        """ delete_connection(id:int) -> None 
        Delete connection identified by id from pipeline.
           
        """

        if not self.has_connection_with_id(id):
            raise VistrailsInternalError("id %s missing in connections" % id)
        conn = self.connections[id]
        # self.connections.pop(id)
        self.db_delete_object(id, 'connection')
        if conn.source is not None and conn.destination is not None and \
                (conn.destinationId, conn.id) in \
                self.graph.edges_from(conn.sourceId):
            self.graph.delete_edge(conn.sourceId, conn.destinationId, conn.id)

            c = conn
            source_name = c.source.name
            output_ports = self.modules[c.sourceId].connected_output_ports
            output_ports[source_name] -= 1
                
            dest_name = c.destination.name
            input_ports = self.modules[c.destinationId].connected_input_ports
            input_ports[dest_name] -= 1

        if id in self._connection_signatures:
            del self._connection_signatures[id]
        
    def add_parameter(self, param, parent_type, parent_id):
        self.db_add_object(param, parent_type, parent_id)
        if not self.has_alias(param.alias):
            self.change_alias(param.alias, 
                              param.vtType, 
                              param.real_id,
                              parent_type,
                              parent_id,
                              None)

    def delete_parameter(self, param_id, param_type, parent_type, parent_id):
        self.db_delete_object(param_id, ModuleParam.vtType,
                              parent_type, parent_id)
        self.remove_alias(ModuleParam.vtType, param_id, parent_type, 
                          parent_id, None)

    def change_parameter(self, old_param_id, param, parent_type, parent_id):
        self.remove_alias(ModuleParam.vtType, old_param_id, 
                          parent_type, parent_id, None)
        self.db_change_object(old_param_id, param,
                              parent_type, parent_id)
        if not self.has_alias(param.alias):
            self.change_alias(param.alias, 
                              param.vtType, 
                              param.real_id,
                              parent_type,
                              parent_id,
                              None)

    def add_port(self, port, parent_type, parent_id):
        self.db_add_object(port, parent_type, parent_id)
        connection = self.connections[parent_id]
        if connection.source is not None and \
                connection.destination is not None:
            self.graph.add_edge(connection.sourceId, 
                                connection.destinationId, 
                                connection.id)
            c = connection
            source_name = c.source.name
            output_ports = self.modules[c.sourceId].connected_output_ports
            if source_name not in output_ports:
                output_ports[source_name] = 0
            output_ports[source_name] += 1
                
            dest_name = c.destination.name
            input_ports = self.modules[c.destinationId].connected_input_ports
            if dest_name not in input_ports:
                input_ports[dest_name] = 0
            input_ports[dest_name] += 1

    def delete_port(self, port_id, port_type, parent_type, parent_id):
        conn = self.connections[parent_id]
        if len(conn.ports) >= 2:
            self.graph.delete_edge(conn.sourceId, 
                                   conn.destinationId, 
                                   conn.id)
            c = conn
            source_name = c.source.name
            output_ports = self.modules[c.sourceId].connected_output_ports
            output_ports[source_name] -= 1
                
            dest_name = c.destination.name
            input_ports = self.modules[c.destinationId].connected_input_ports
            input_ports[dest_name] -= 1
            
        self.db_delete_object(port_id, Port.vtType, parent_type, parent_id)

    def change_port(self, old_port_id, port, parent_type, parent_id):
        connection = self.connections[parent_id]
        if len(connection.ports) >= 2:
            source_list = self.graph.adjacency_list[connection.sourceId]
            source_list.remove((connection.destinationId, connection.id))
            dest_list = \
                self.graph.inverse_adjacency_list[connection.destinationId]
            dest_list.remove((connection.sourceId, connection.id))
        self.db_change_object(old_port_id, port, parent_type, parent_id)
        if len(connection.ports) >= 2:
            source_list = self.graph.adjacency_list[connection.sourceId]
            source_list.append((connection.destinationId, connection.id))
            dest_list = \
                self.graph.inverse_adjacency_list[connection.destinationId]
            dest_list.append((connection.sourceId, connection.id))

    def add_port_to_registry(self, portSpec, moduleId):
        m = self.get_module_by_id(moduleId)
        m.add_port_spec(portSpec)

    def add_portSpec(self, port_spec, parent_type, parent_id):
        # self.db_add_object(port_spec, parent_type, parent_id)
        self.add_port_to_registry(port_spec, parent_id)
        
    def delete_port_from_registry(self, id, moduleId):
        m = self.get_module_by_id(moduleId)
        portSpec = m.port_specs[id]
        m.delete_port_spec(portSpec)

    def delete_portSpec(self, spec_id, portSpec_type, parent_type, parent_id):
        self.delete_port_from_registry(spec_id, parent_id)
        # self.db_delete_object(spec_id, PortSpec.vtType, parent_type, parent_id)

    def change_portSpec(self, old_spec_id, port_spec, parent_type, parent_id):
        self.delete_port_from_registry(old_spec_id, parent_id)
        # self.db_change_object(old_spec_id, port_spec, parent_type, parent_id)
        self.add_port_to_registry(port_spec, parent_id)

    def add_alias(self, name, type, oId, parentType, parentId, mId):
        """add_alias(name: str, oId: int, parentType:str, parentId: int, 
                     mId: int) -> None 
        Add alias to pipeline
          
        """
        if self.has_alias(name):
            raise VistrailsInternalError("duplicate alias")
        if mId is not None:
            self.aliases[name] = (type, oId, parentType, parentId, mId)
        else:
            mid = None
            for _mod in self.modules.itervalues():
                for _fun in _mod.functions:
                    for _par in _fun.parameters:
                        if (_par.vtType == type and _par.real_id == oId and
                            _fun.vtType == parentType and 
                            _fun.real_id == parentId):
                            mid = _mod.id
                            break
            if mid is not None:
                self.aliases[name] = (type, oId, parentType, parentId, mid)
                
    def remove_alias_by_name(self, name):
        """remove_alias_by_name(name: str) -> None
        Remove alias with given name """
        if self.has_alias(name):
            del self.aliases[name]

    def remove_alias(self, type, oId, parentType, parentId, mId):
        """remove_alias(name: str, type:str, oId: int, parentType: str, 
                        parentId: int, mId: int)-> None
        Remove alias identified by oId """
        if mId is not None:
            try:
                oldname = self.aliases.inverse[(type,oId, parentType, parentId, mId)]
                del self.aliases[oldname]
            except KeyError:
                pass
        else:
            oldname = None
            for aname,(t,o,pt,pid,mid) in self.aliases.iteritems():
                if (t == type and o == oId and pt == parentType and 
                    pid == parentId):
                    oldname = aname
                    break
            if oldname:
                del self.aliases[oldname]

    def change_alias(self, name, type, oId, parentType, parentId, mId):
        """change_alias(name: str, type:str oId:int, parentType:str,
                        parentId:int, mId: int)-> None
        Change alias if name is non empty. Else remove alias
        
        """
        if name == "":
            self.remove_alias(type, oId, parentType, parentId, mId)
        else:
            if not self.has_alias(name):
                self.remove_alias(type, oId, parentType, parentId, mId)
                self.add_alias(name, type, oId, parentType, parentId, mId)
                
    def get_alias_str_value(self, name):
        """ get_alias_str_value(name: str) -> str
        returns the strValue of the parameter with alias name

        """
        try:
            what, oId, parentType, parentId, mId = self.aliases[name]
        except KeyError:
            return ''
        else:
            if what == 'parameter':
                parameter = self.db_get_object(what, oId)
                return parameter.strValue
            else:
                raise VistrailsInternalError("only parameters are supported")

    def set_alias_str_value(self, name, value):
        """ set_alias_str_value(name: str, value: str) -> None
        sets the strValue of the parameter with alias name 
        
        """
        try:
            what, oId, parentType, parentId, mId = self.aliases[name]
        except KeyError:
            pass
        else:
            if what == 'parameter':
                #FIXME: check if a change parameter action needs to be generated
                parameter = self.db_get_object(what, oId)
                parameter.strValue = str(value)
            else:
                raise VistrailsInternalError("only parameters are supported")
        
    def get_module_by_id(self, id):
        """get_module_by_id(id: int) -> Module
        Accessor. id is the Module id.
        
        """
        result = self.modules[id]
        if result.vtType != Abstraction.vtType and \
                result.vtType != Group.vtType and result.package is None:
            registry = get_module_registry()
            debug.critical('module %d is missing package' % id)
            descriptor = registry.get_descriptor_from_name_only(result.name)
            result.package = descriptor.identifier
        return result
    
    def get_connection_by_id(self, id):
        """get_connection_by_id(id: int) -> Connection
        Accessor. id is the Connection id.
        
        """
        self.ensure_connection_specs([id])
        return self.connections[id]
    
    def module_count(self):
        """ module_count() -> int 
        Returns the number of modules in the pipeline.
        
        """
        return len(self.modules)
    
    def connection_count(self):
        """connection_count() -> int 
        Returns the number of connections in the pipeline.
        
        """
        return len(self.connections)
    
    def has_module_with_id(self, id):
        """has_module_with_id(id: int) -> boolean 
        Checks whether given module exists.

        """
        return id in self.modules
    
    def has_connection_with_id(self, id):
        """has_connection_with_id(id: int) -> boolean 
        Checks whether given connection exists.

        """
        return id in self.connections

    def has_alias(self, name):
        """has_alias(name: str) -> boolean 
        Checks whether given alias exists.

        """
        return name in self.aliases

    def out_degree(self, id):
        """out_degree(id: int) -> int - Returns the out-degree of a module. """
        return self.graph.out_degree(id)

    ##########################################################################
    # Caching-related

    # Modules

    def module_signature(self, module_id):
        """module_signature(module_id): string
        Returns the signature for the module with given module_id."""
        try:
            return self._module_signatures[module_id]
        except KeyError:
            registry = get_module_registry()
            m = self.modules[module_id]
            sig = registry.module_signature(self, m)
            self._module_signatures[module_id] = sig
            return sig
    
    def module_id_from_signature(self, signature):
        """module_id_from_signature(sig): int
        Returns the module_id that corresponds to the given signature.
        This must have been previously computed."""
        return self._module_signatures.inverse[signature]

    def has_module_signature(self, signature):
        return signature in self._module_signatures.inverse

    # Subpipelines

    def subpipeline_signature(self, module_id, visited_ids=None):
        """subpipeline_signature(module_id): string
        Returns the signature for the subpipeline whose sink id is module_id."""
        if visited_ids is None:
            visited_ids = set([module_id])
        elif module_id in visited_ids:
            raise CycleInPipeline()
        try:
            return self._subpipeline_signatures[module_id]
        except KeyError:
            upstream_sigs = [(self.subpipeline_signature(
                                      m,
                                      visited_ids | set([module_id])) +
                              Hasher.connection_signature(
                                      self.connections[edge_id]))
                             for (m, edge_id) in
                             self.graph.edges_to(module_id)]
            module_sig = self.module_signature(module_id)
            sig = Hasher.subpipeline_signature(module_sig,
                                               upstream_sigs)
            self._subpipeline_signatures[module_id] = sig
            return sig

    def subpipeline_id_from_signature(self, signature):
        """subpipeline_id_from_signature(sig): int
        Returns the module_id that corresponds to the given signature.
        This must have been previously computed."""
        return self._subpipeline_signatures.inverse[signature]

    def has_subpipeline_signature(self, signature):
        return signature in self._subpipeline_signatures.inverse

    # Connections

    def connection_signature(self, connection_id):
        """connection_signature(id): string
        Returns the signature for the connection with given id."""
        try:
            return self._connection_signatures[connection_id]
        except KeyError:
            c = self.connections[connection_id]
            source_sig = self.subpipeline_signature(c.sourceId)
            dest_sig = self.subpipeline_signature(c.destinationId)
            sig = Hasher.connection_subpipeline_signature(c, source_sig,
                                                          dest_sig)
            self._connection_signatures[connection_id] = sig
            return sig

    def connection_id_from_signature(self, signature):
        return self._connection_signatures.inverse[signature]

    def has_connection_signature(self, signature):
        return signature in self._connection_signatures.inverse

    def refresh_signatures(self):
        self._connection_signatures = {}
        self._subpipeline_signatures = {}
        self._module_signatures = {}
        self.compute_signatures()

    def compute_signatures(self):
        """compute_signatures(): compute all module and subpipeline signatures
        for this pipeline."""
        for i in self.modules.iterkeys():
            self.subpipeline_signature(i)
        for c in self.connections.iterkeys():
            self.connection_signature(c)

    ##########################################################################
    # Registry-related

    def validate(self, raise_exception=True, vistrail_vars={}):
        # want to check entire pipeline and reconcile it with the
        # registry - if anything fails, generate invalid pipeline with
        # the errors
        exceptions = set()
        try:
            self.ensure_modules_are_on_registry()
        except InvalidPipeline, e:
            exceptions.update(e.get_exception_set())

        # do this before we check connection specs because it is
        # possible that a subpipeline invalidates the module, meaning
        # we shouldn't check the connection specs
        for module in self.modules.itervalues():
            if module.is_valid and (module.is_group() or 
                                    module.is_abstraction()):
                try:
                    subpipeline = module.pipeline
                    if subpipeline is not None:
                        subpipeline.validate()
                except InvalidPipeline, e:
                    module.is_valid = False
                    e._module_id = module.id
                    exceptions.add(e)
                if module.is_abstraction():
                    try:
                        desc = module.module_descriptor
                        if long(module.internal_version) != long(desc.version):
                            exceptions.add(MissingModuleVersion(desc.package, desc.name, desc.namespace, desc.version, desc.package_version, module.id))
                    except Exception:
                        pass
        try:
            self.ensure_port_specs()
        except InvalidPipeline, e:
            exceptions.update(e.get_exception_set())
        try:
            self.ensure_connection_specs()
        except InvalidPipeline, e:
            exceptions.update(e.get_exception_set())
        try:
            self.ensure_functions()
        except InvalidPipeline, e:
            exceptions.update(e.get_exception_set())
        try:
            self.ensure_vistrail_variables(vistrail_vars)
        except InvalidPipeline, e:
            exceptions.update(e.get_exception_set())
        
        self.check_subworkflow_versions()
        
        if len(exceptions) > 0:
            if raise_exception:
                raise InvalidPipeline(exceptions, self)
            else:
                self.is_valid = False
                return False

        self.mark_list_depth()

        self.is_valid = True
        return True

    def ensure_old_modules_have_package_names(self):
        """ensure_old_modules_have_package_names()

        Makes sure each module has a package associated with it.

        """
        for i in self.modules.iterkeys():
            self.get_module_by_id(i)

    def ensure_connection_specs(self, connection_ids=None):
        """ensure_connection_specs(connection_ids=None) -> None.

        Computes the specs for the connections in connection_ids. If
        connection_ids is None, computes it for every connection in the pipeline.
        """
        exceptions = set()

        # print 'ensure_connection_specs:', sorted(self.modules.keys())

        def find_spec(port):
            port.is_valid = False
            module = self.get_module_by_id(port.moduleId)
            port_type_map = PortSpec.port_type_map
            try:
                # print 'running get_port_spec', port.name
                port.spec = module.get_port_spec(port.name, 
                                            port_type_map.inverse[port.type])
                # print 'got spec', spec, spec.sigstring
            except ModuleRegistryException, e:
                # debug.critical('CONNECTION EXCEPTION: %s' % e)
                exceptions.add(e)
            else:
                if port.spec.is_valid:
                    port.is_valid = True
            
        if connection_ids is None:
            connection_ids = self.connections.iterkeys()
        for conn_id in connection_ids:
            conn = self.connections[conn_id]
            # print 'checking connection', conn_id, conn.source.moduleId, conn.source.moduleName, conn.source.name, conn.destination.moduleId, conn.destination.moduleName, conn.destination.name
            src_module = self.modules[conn.source.moduleId]
            if src_module.is_valid:
                # print 'src_module:', src_module.name, src_module.id
                find_spec(conn.source)
            
            dst_module = self.modules[conn.destination.moduleId]
            if dst_module.is_valid:
                # print 'dst_module:', dst_module.name, dst_module.id
                find_spec(conn.destination)

            # if not conn.source.spec:
            # conn.source.spec = find_spec(conn.source)
            # if not conn.destination.spec:
            # conn.destination.spec = find_spec(conn.destination)
            # print 'source spec:', conn.source.spec.sigstring
            # print 'dest spec:', conn.destination.spec.sigstring

        if len(exceptions) > 0:
            raise InvalidPipeline(exceptions, self)

    def ensure_modules_are_on_registry(self, module_ids=None):
        """ensure_modules_are_on_registry(module_ids: optional list of module ids) -> None

        Queries the module registry for the module information in the
        given modules.  The only goal of this function is to trigger
        exceptions in the registry that will be treated somewhere else
        in the calling stack.
        
        If modules are not on registry, the registry will raise
        ModuleRegistryException exceptions that should be caught and handled.

        if no module_ids list is given, we assume every module in the pipeline.
        """
        def find_descriptors(pipeline, module_ids=None):
            registry = get_module_registry()
            conf = get_vistrails_configuration()
            if module_ids is None:
                module_ids = pipeline.modules.iterkeys()
            exceptions = set()
            for mid in module_ids:
                module = pipeline.modules[mid]
                module.is_valid = False
                if not module.version:
                    module.version = '0'
                try:
                    # FIXME check for upgrades, otherwise use similar
                    # descriptor, the old behavior
                    descriptor = module.module_descriptor
                except ModuleRegistryException, e:
                    e._module_id = mid
                    exceptions.add(e)
                else:
                    module.is_valid = True
            return exceptions
        # end find_descriptors

        exceptions = find_descriptors(self, module_ids)
        if len(exceptions) > 0:
            raise InvalidPipeline(exceptions, self)

    def ensure_functions(self):
        exceptions = set()
        reg = get_module_registry()
        for module in self.modules.itervalues():
            for function in module.functions:
                is_valid = True
                if module.is_valid and not module.has_port_spec(function.name, 
                                                                'input'):
                    is_valid = False
                    e = MissingFunction(function.name, module.name, module.id)
                    e._module_id = module.id
                    exceptions.add(e)
                pos_map = {}
                for p in function.parameters:
                    if p.identifier == '':
                        idn = get_vistrails_basic_pkg_id()
                    else:
                        idn = p.identifier

                    try:
                        desc = reg.get_module_by_name(idn,
                                                      p.type,
                                                      p.namespace)
                    except ModuleRegistryException, e:
                        is_valid = False
                        e._module_id = module.id
                        exceptions.add(e)

                    if p.pos in pos_map:
                        is_valid = False
                        e = VistrailsInternalError("Module %d has multiple "
                                                   "values for parameter %d "
                                                   "of function %s (%d)" % \
                                                       (module.id,
                                                        p.pos,
                                                        function.name,
                                                        function.real_id))
                        exceptions.add(e)
                    pos_map[p.pos] = p
                function.is_valid = is_valid
        if len(exceptions) > 0:
            raise InvalidPipeline(exceptions, self)
        
    def ensure_vistrail_variables(self, vistrail_vars):
        var_uuids = [var.uuid for var in vistrail_vars]
        exceptions = set()
        for module in self.modules.itervalues():
            if module.is_vistrail_var():
                # first check if value is already set
                # (used by parameter explorations)
                value_set = False
                for func in module.functions:
                    if func.name == 'value':
                        if func.params[0].strValue:
                            value_set = True
                            continue
                if value_set:
                    continue
                var_uuid = module.get_vistrail_var()
                if var_uuid not in var_uuids:
                    e = MissingVistrailVariable(var_uuid, module.package, 
                                                module.name, module.namespace)
                    exceptions.add(e)
        if len(exceptions) > 0:
            raise InvalidPipeline(exceptions, self)

    def ensure_port_specs(self):
        exceptions = set()
        for module in self.modules.itervalues():
            # if module.is_valid:
            try:
                for port_spec in module.port_specs.itervalues():
                    try:
                        port_spec.descriptors()
                    except MissingPackage, e:
                        port_spec.is_valid = False
                        e._module_id = module.id
                        exceptions.add(e)
                    except ModuleRegistryException, e:
                        e = PortMismatch(module.package, module.name,
                                         module.namespace, port_spec.name,
                                         port_spec.type, port_spec.sigstring)
                        port_spec.is_valid = False
                        e._module_id = module.id
                        exceptions.add(e)
            except ModuleRegistryException, e:
                if module.is_valid:
                    module.is_valid = False
    
        if len(exceptions) > 0:
            raise InvalidPipeline(exceptions, self)

    def check_subworkflow_versions(self):
        reg = get_module_registry()
        for module in self.modules.itervalues():
            if module.is_valid and module.is_abstraction():
                module.check_latest_version()

    def mark_list_depth(self, module_ids=None):
        """mark_list_depth() -> list

        Updates list_depth variable on each module according to list depth of
        connecting port specs. This decides at what list depth the module
        needs to be executed.
        List ports have default depth 1

        module_ids: list of module_ids - The id:s of modules that have changed
        All modules upstream of these will be skipped, since markings only
        affects downstream. This slightly increases performance.

        """
        result = []
        is_upstream = module_ids
        for module_id in self.graph.vertices_topological_sort():
            if is_upstream:
                if module_id in module_ids:
                    is_upstream = False
                else:
                    continue
            module = self.get_module_by_id(module_id)
            module.list_depth = 0
            ports = []
            for module_from_id, conn_id in self.graph.edges_to(module_id):
                prev_depth = self.get_module_by_id(module_from_id).list_depth
                conn = self.get_connection_by_id(conn_id)
                source_depth = 0
                from vistrails.core.modules.basic_modules import List, Variant
                if conn.source.spec:
                    source_depth = conn.source.spec.depth
                    src_descs = conn.source.spec.descriptors()
                    # Lists have depth 1 if dest has depth>1 or is a list
                    if len(src_descs) == 1 and src_descs[0].module == List:
                        source_depth += 1
                dest_depth = 0
                if conn.destination.spec:
                    dest_depth = conn.destination.spec.depth
                    dest_descs = conn.destination.spec.descriptors()
                    # Lists have depth 1
                    if len(dest_descs) == 1 and dest_descs[0].module == List:
                        dest_depth += 1
                    # special case: if src is List and dst is Variant
                    # we should treat the Variant as having depth 0
                    if conn.source.spec:
                        if len(src_descs)==1 and src_descs[0].module == List and \
                           len(dest_descs)==1 and dest_descs[0].module == Variant:
                            source_depth -= 1
                        if len(src_descs)==1 and src_descs[0].module == Variant and \
                           len(dest_descs)==1 and dest_descs[0].module == List:
                            dest_depth -= 1
                depth = prev_depth + source_depth - dest_depth
                if depth > 0 and conn.destination.spec.name not in ports:
                    ports.append(conn.destination.spec.name)
                # if dest depth is greater the input will be wrapped in a
                # list to match its depth
                # if source depth is greater this module will be executed
                # once for each input in the (possibly nested) list
                module.list_depth = max(module.list_depth, depth)
            result.append((module_id, module.list_depth))
            module.iterated_ports = ports
        return result


    ##########################################################################
    # Debugging

    def show_comparison(self, other):
        if type(other) != type(self):
            print "type mismatch"
            return
        if len(self.module_list) != len(other.module_list):
            print "module lists of different sizes"
            return
        if len(self.connection_list) != len(other.connection_list):
            print "Connection lists of different sizes"
            return
        for m_id, m in self.modules.iteritems():
            if not m_id in other.modules:
                print "module %d in self but not in other" % m_id
                return
            if m <> other.modules[m_id]:
                print "module %s in self doesn't match module %s in other" % (m,  other.modules[m_id])
                return
        for m_id, m in other.modules.iteritems():
            if not m_id in self.modules:
                print "module %d in other but not in self" % m_id
                return
            # no need to check equality since this was already done before
        for c_id, c in self.connections.iteritems():
            if not c_id in other.connections:
                print "connection %d in self but not in other" % c_id
                return
            if c <> other.connections[c_id]:
                print "connection %s in self doesn't match connection %s in other" % (c,  other.connections[c_id])
                return
        for c_id, c, in other.connections.iteritems():
            if not c_id in self.connections:
                print "connection %d in other but not in self" % c_id
                return
            # no need to check equality since this was already done before
        assert self == other

    ##########################################################################
    # Operators

    def __ne__(self, other):
        return not self.__eq__(other)

# There's a bug in this code that's not easily worked around: if
# modules are in different order in the list, there's no easy way to
# check for equality. The solution is to move to a check that
# takes module and connection ids into account.
#     def __eq__(self, other):
#         if type(other) != type(self):
#             return False
#         if len(self.module_list) != len(other.module_list):
#             return False
#         if len(self.connection_list) != len(other.connection_list):
#             return False
#         for f, g in zip(self.module_list, other.module_list):
#             if f != g:
#                 return False
#         for f, g in zip(self.connection_list, other.connection_list):
#             if f != g:
#                 return False
#         return True

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        if len(self.module_list) != len(other.module_list):
            return False
        if len(self.connection_list) != len(other.connection_list):
            return False
        for m_id, m in self.modules.iteritems():
            if not m_id in other.modules:
                return False
            if m <> other.modules[m_id]:
                return False
        for m_id, m in other.modules.iteritems():
            if not m_id in self.modules:
                return False
            # no need to check equality since this was already done before
        for c_id, c in self.connections.iteritems():
            if not c_id in other.connections:
                return False
            if c <> other.connections[c_id]:
                return False
        for c_id, c, in other.connections.iteritems():
            if not c_id in self.connections:
                return False
            # no need to check equality since this was already done before
        return True

    def __str__(self):
        return ("(Pipeline Modules: %s Graph:%s)@%X" %
                ([(m, str(v)) for (m,v) in sorted(self.modules.items())],
                 str(self.graph),
                 id(self)))


################################################################################


class TestPipeline(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # make sure pythonCalc is loaded
        from vistrails.core.packagemanager import get_package_manager
        pm = get_package_manager()
        if 'pythonCalc' not in pm._package_list: # pragma: no cover # pragma: no branch
            pm.late_enable_package('pythonCalc')

    def create_default_pipeline(self, id_scope=None):
        if id_scope is None:
            id_scope = IdScope()
        
        p = Pipeline()
        p.id = id_scope.getNewId(Pipeline.vtType)

        def module1(p):
            def f1():
                f = ModuleFunction()
                f.real_id = id_scope.getNewId(ModuleFunction.vtType)
                f.name = 'op'
                f.returnType = 'void'
                param = ModuleParam()
                param.type = 'String'
                param.strValue = '+'
                f.params.append(param)
                return f
            def f2():
                f = ModuleFunction()
                f.real_id = id_scope.getNewId(ModuleFunction.vtType)
                f.name = 'value1'
                f.returnType = 'void'
                param = ModuleParam()
                param.type = 'Float'
                param.strValue = '2.0'
                f.params.append(param)
                return f
            def f3():
                f = ModuleFunction()
                f.real_id = id_scope.getNewId(ModuleFunction.vtType)
                f.name = 'value2'
                f.returnType = 'void'
                param = ModuleParam()
                param.type = 'Float'
                param.strValue = '4.0'
                f.params.append(param)
                return f
            def cp1():
                f = ModuleControlParam()
                f.id = id_scope.getNewId(ModuleControlParam.vtType)
                f.name = 'cpname1'
                f.value = 'cpvalue[]'
                return f
            m = Module()
            m.id = id_scope.getNewId(Module.vtType)
            m.name = 'PythonCalc'
            m.package = 'org.vistrails.vistrails.pythoncalc'
            m.functions.append(f1())
            m.control_parameters.append(cp1())
            return m
        
        def module2(p):
            def f1():
                f = ModuleFunction()
                f.real_id = id_scope.getNewId(ModuleFunction.vtType)
                f.name = 'op'
                f.returnType = 'void'
                param = ModuleParam()
                param.type = 'String'
                param.strValue = '+'
                f.params.append(param)
                return f
            m = Module()
            m.id = id_scope.getNewId(Module.vtType)
            m.name = 'PythonCalc'
            m.package = 'org.vistrails.vistrails.pythoncalc'
            m.functions.append(f1())
            return m
        m1 = module1(p)
        p.add_module(m1)
        m2 = module1(p)
        p.add_module(m2)
        m3 = module2(p)
        p.add_module(m3)

        c1 = Connection()
        c1.sourceId = m1.id
        c1.destinationId = m3.id
        c1.source.id = id_scope.getNewId(Port.vtType)
        c1.destination.id = id_scope.getNewId(Port.vtType)
        c1.source.name = 'value'
        c1.source.moduleName = 'PythonCalc'
        c1.destination.name = 'value1'
        c1.destination.moduleName = 'PythonCalc'
        c1.id = id_scope.getNewId(Connection.vtType)
        p.add_connection(c1)

        c2 = Connection()
        c2.sourceId = m2.id
        c2.destinationId = m3.id
        c2.source.id = id_scope.getNewId(Port.vtType)
        c2.destination.id = id_scope.getNewId(Port.vtType)
        c2.source.name = 'value'
        c2.source.moduleName = 'PythonCalc'
        c2.destination.name = 'value2'
        c2.destination.moduleName = 'PythonCalc'
        c2.id = id_scope.getNewId(Connection.vtType)

        p.add_connection(c2)
        p.compute_signatures()
        return p

    def create_pipeline2(self, id_scope=None):
        basic_pkg = get_vistrails_basic_pkg_id()
        if id_scope is None:
            id_scope = IdScope(remap={Abstraction.vtType: Module.vtType})

        param1 = ModuleParam(id=id_scope.getNewId(ModuleParam.vtType),
                             type='Int',
                             val='1')
        param2 = ModuleParam(id=id_scope.getNewId(ModuleParam.vtType),
                             type='Float',
                             val='1.3456')
        func1 = ModuleFunction(id=id_scope.getNewId(ModuleFunction.vtType),
                               name='value',
                               parameters=[param1])
        func2 = ModuleFunction(id=id_scope.getNewId(ModuleFunction.vtType),
                               name='floatVal',
                               parameters=[param2])
        loc1 = Location(id=id_scope.getNewId(Location.vtType),
                        x=12.342,
                        y=-19.432)
        loc2 = Location(id=id_scope.getNewId(Location.vtType),
                        x=21.34,
                        y=456.234)
        m1 = Module(id=id_scope.getNewId(Module.vtType),
                    package=basic_pkg,
                    name='String',
                    location=loc1,
                    functions=[func1])
        m2 = Abstraction(id=id_scope.getNewId(Abstraction.vtType),
                         internal_version=13,
                         location=loc2,
                         functions=[func2])
        source = Port(id=id_scope.getNewId(Port.vtType),
                      type='source', 
                      moduleId=m1.id, 
                      moduleName='String', 
                      name='value',
                      signature='(%s:String)' % basic_pkg)
        destination = Port(id=id_scope.getNewId(Port.vtType),
                           type='destination',
                           moduleId=m2.id,
                           moduleName='Abstraction',
                           name='self',
                           signature='()')
        c1 = Connection(id=id_scope.getNewId(Connection.vtType),
                        ports=[source, destination])
        pipeline = Pipeline(id=id_scope.getNewId(Pipeline.vtType),
                            modules=[m1, m2],
                            connections=[c1])
        return pipeline

    def setUp(self):
        self.pipeline = self.create_default_pipeline()
        self.sink_id = 2

    def test_create_pipeline_signature(self):
        self.pipeline.subpipeline_signature(self.sink_id)

    def test_delete_signatures(self):
        """Makes sure signatures are deleted when other things are."""
        p = self.create_default_pipeline()
        m_sig_size_before = len(p._module_signatures)
        c_sig_size_before = len(p._connection_signatures)
        p_sig_size_before = len(p._subpipeline_signatures)
        p.delete_connection(0)
        p.delete_module(0)
        m_sig_size_after = len(p._module_signatures)
        c_sig_size_after = len(p._connection_signatures)
        p_sig_size_after = len(p._subpipeline_signatures)
        self.assertNotEquals(m_sig_size_before, m_sig_size_after)
        self.assertNotEquals(c_sig_size_before, c_sig_size_after)
        self.assertNotEquals(p_sig_size_before, p_sig_size_after)

    def test_delete_connections(self):
        p = self.create_default_pipeline()
        p.delete_connection(0)
        p.delete_connection(1)
        p.delete_module(2)
        self.assertEquals(len(p.connections), 0)

    def test_basic(self):
        """Makes sure pipeline can be created, modules and connections
        can be added and deleted."""
        p = self.create_default_pipeline()
    
    def test_copy(self):
        id_scope = IdScope()
        
        p1 = self.create_default_pipeline(id_scope)
        p2 = copy.copy(p1)
        self.assertEquals(p1, p2)
        self.assertEquals(p1.id, p2.id)
        p3 = p1.do_copy(True, id_scope, {})
        self.assertNotEquals(p1, p3)
        self.assertNotEquals(p1.id, p3.id)

    def test_copy2(self):
        # need to id modules and abstraction_modules with same counter
        id_scope = IdScope(remap={Abstraction.vtType: Module.vtType})
        
        p1 = self.create_pipeline2(id_scope)
        p2 = copy.copy(p1)
        self.assertEquals(p1, p2)
        self.assertEquals(p1.id, p2.id)
        p3 = p1.do_copy(True, id_scope, {})
        self.assertNotEquals(p1, p3)
        self.assertNotEquals(p1.id, p3.id)

    def test_serialization(self):
        import vistrails.core.db.io
        p1 = self.create_default_pipeline()
        xml_str = vistrails.core.db.io.serialize(p1)
        p2 = vistrails.core.db.io.unserialize(xml_str, Pipeline)
        self.assertEquals(p1, p2)
        self.assertEquals(p1.id, p2.id)        

    def test_serialization2(self):
        import vistrails.core.db.io
        p1 = self.create_pipeline2()
        xml_str = vistrails.core.db.io.serialize(p1)
        p2 = vistrails.core.db.io.unserialize(xml_str, Pipeline)
        self.assertEquals(p1, p2)
        self.assertEquals(p1.id, p2.id)        

    def test_aliases(self):
        """ Exercises aliases manipulation """
        import vistrails.core.db.action
        from vistrails.core.db.locator import XMLFileLocator
        import vistrails.core.system
        v = XMLFileLocator( \
            vistrails.core.system.vistrails_root_directory() +
            '/tests/resources/test_alias.xml').load()

        p1 = v.getPipeline('alias')
        p2 = v.getPipeline('alias')
        
        # testing removing an alias
        old_param = p1.modules[0].functions[0].params[0]
        func = p1.modules[0].functions[0]
        #old_id = p1.modules[0].functions[0].params[0].db_id
        #old_f_id = p1.modules[0].functions[0].db_id
        new_param = ModuleParam(id=-1,
                                pos=old_param.pos,
                                name=old_param.name,
                                alias="",
                                val=str(1.0),
                                type=old_param.type)
        action_spec = ('change', old_param, new_param,
                       func.vtType, func.real_id)
        action = vistrails.core.db.action.create_action([action_spec])
        p1.perform_action(action)
        self.assertEquals(p1.has_alias('v1'),False)
        v1 = p2.aliases['v1']
        old_param2 = p2.modules[0].functions[0].params[0]
        new_param2 = ModuleParam(id=old_param.real_id,
                                pos=old_param.pos,
                                name=old_param.name,
                                alias="v1",
                                val=str(v),
                                type=old_param.type)
        func2 = p2.modules[0].functions[0]
        action2 = vistrails.core.db.action.create_action([('change',
                                                 old_param2,
                                                 new_param2,
                                                 func2.vtType,
                                                 func2.real_id)
                                                ])
        p2.perform_action(action2)
        self.assertEquals(v1, p2.aliases['v1'])
        
    def test_module_signature(self):
        """Tests signatures for modules with similar (but not equal)
        parameter specs."""
        pycalc_pkg = 'org.vistrails.vistrails.pythoncalc'
        p1 = Pipeline()
        p1_functions = [ModuleFunction(name='value1',
                                       parameters=[ModuleParam(type='Float',
                                                               val='1.0',
                                                               )],
                                       ),
                        ModuleFunction(name='value2',
                                       parameters=[ModuleParam(type='Float',
                                                               val='2.0',
                                                               )],
                                       )]
        p1.add_module(Module(name='PythonCalc',
                             package=pycalc_pkg,
                             id=3,
                             functions=p1_functions))

        p2 = Pipeline()
        p2_functions = [ModuleFunction(name='value1',
                                       parameters=[ModuleParam(type='Float',
                                                               val='2.0',
                                                               )],
                                       ),
                        ModuleFunction(name='value2',
                                       parameters=[ModuleParam(type='Float',
                                                               val='1.0',
                                                               )],
                                       )]
        p2.add_module(Module(name='PythonCalc', 
                             package=pycalc_pkg,
                             id=3,
                             functions=p2_functions))

        self.assertNotEquals(p1.module_signature(3),
                             p2.module_signature(3))

    def test_find_method(self):
        p1 = Pipeline()
        p1_functions = [ModuleFunction(name='i1',
                                       parameters=[ModuleParam(type='Float',
                                                               val='1.0',
                                                               )],
                                       ),
                        ModuleFunction(name='i2',
                                       parameters=[ModuleParam(type='Float',
                                                               val='2.0',
                                                               )],
                                       )]
        p1.add_module(Module(name='CacheBug', 
                            package='org.vistrails.vistrails.console_mode_test',
                            id=3,
                            functions=p1_functions))

        self.assertEquals(p1.find_method(3, 'i1'), 0)
        self.assertEquals(p1.find_method(3, 'i2'), 1)
        self.assertEquals(p1.find_method(3, 'i3'), -1)
        self.assertRaises(KeyError, p1.find_method, 4, 'i1')

    def test_str(self):
        p1 = Pipeline()
        p1_functions = [ModuleFunction(name='i1',
                                       parameters=[ModuleParam(type='Float',
                                                               val='1.0',
                                                               )],
                                       ),
                        ModuleFunction(name='i2',
                                       parameters=[ModuleParam(type='Float',
                                                               val='2.0',
                                                               )],
                                       )]
        p1.add_module(Module(name='CacheBug', 
                            package='org.vistrails.vistrails.console_mode_test',
                            id=3,
                            functions=p1_functions))
        str(p1)

    def test_pipeline_equality_module_list_out_of_order(self):
        p1 = Pipeline()
        p1.add_module(Module(name='Foo',
                             package='bar',
                             id=0,
                             functions=[]))
        p1.add_module(Module(name='Foo2',
                             package='bar',
                             id=1,
                             functions=[]))
        p2 = Pipeline()
        p2.add_module(Module(name='Foo2',
                             package='bar',
                             id=1,
                             functions=[]))
        p2.add_module(Module(name='Foo',
                             package='bar',
                             id=0,
                             functions=[]))
        assert p1 == p2

    def test_incorrect_port_spec(self):
        import vistrails.core.modules.basic_modules
        p = Pipeline()
        basic_version = vistrails.core.modules.basic_modules.version
        basic_pkg = vistrails.core.modules.basic_modules.identifier
        m1 = Module(name="String",
                    package=basic_pkg,
                    version=basic_version,
                    id=1L)
        m2 = Module(name="String",
                    package=basic_pkg,
                    version=basic_version,
                    id=2L)
        source = Port(id=1L,
                      type='source', 
                      moduleId=m1.id, 
                      moduleName='String', 
                      name='value',
                      signature='(%s:StringBean)' % basic_pkg)
        destination = Port(id=2L,
                           type='destination',
                           moduleId=m2.id,
                           moduleName='String',
                           name='value',
                           signature='(%s:NString)' % basic_pkg)
        c1 = Connection(id=1L,
                        ports=[source, destination])
        p.add_module(m1)
        p.add_module(m2)
        p.add_connection(c1)
        p.ensure_modules_are_on_registry()
        p.ensure_connection_specs()
        p_source = p.connections[c1.id].source
        p_destination = p.connections[c1.id].destination
        self.assertEqual(p_source.signature, '(%s:String)' % basic_pkg)
        self.assertEqual(len(p_source.descriptors()), 1)
        self.assertEqual(p_destination.signature, '(%s:String)' % basic_pkg)
        self.assertEqual(len(p_destination.descriptors()), 1)

if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file contains widgets related to the module palette
displaying a list of all modules in Vistrails

QModulePalette
QModuleTreeWidget
QModuleTreeWidgetItem
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core import get_vistrails_application
from vistrails.core import debug
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.system import systemType
from vistrails.core.utils import VistrailsInternalError
from vistrails.gui.common_widgets import (QSearchTreeWindow,
                                QSearchTreeWidget)
from vistrails.gui.module_documentation import QModuleDocumentation
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface

################################################################################

class QModulePalette(QSearchTreeWindow, QVistrailsPaletteInterface):
    """
    QModulePalette just inherits from QSearchTreeWindow to have its
    own type of tree widget

    """
    def __init__(self, parent=None):
        QSearchTreeWindow.__init__(self, parent)
        self.setContentsMargins(0,5,0,0)
        self.packages = {}
        self.namespaces = {}
        self.addButtonsToToolBar()

    def addButtonsToToolBar(self):
        self.expandAction = QtGui.QAction(CurrentTheme.EXPAND_ALL_ICON,
                                           "Expand All", self.toolWindow().toolbar,
                                           triggered=self.expandAll)

        self.collapseAction = QtGui.QAction(CurrentTheme.COLLAPSE_ALL_ICON,
                                           "Collapse All", self.toolWindow().toolbar,
                                           triggered=self.collapseAll)
        self.toolWindow().toolbar.insertAction(self.toolWindow().pinAction,
                                               self.collapseAction)
        self.toolWindow().toolbar.insertAction(self.toolWindow().pinAction,
                                               self.expandAction)

    def expandAll(self):
        self.treeWidget.expandAll()

    def collapseAll(self):
        self.treeWidget.collapseAll()

    def link_registry(self):
        self.updateFromModuleRegistry()
        self.connect_registry_signals()


    def connect_registry_signals(self):
        app = get_vistrails_application()
        app.register_notification("reg_new_module", self.newModule)
        app.register_notification("reg_new_package", self.newPackage)
        app.register_notification("reg_deleted_module", self.deletedModule)
        app.register_notification("reg_deleted_package", self.deletedPackage)
        app.register_notification("reg_show_module", self.showModule)
        app.register_notification("reg_hide_module", self.hideModule)
        app.register_notification("reg_module_updated", self.switchDescriptors)

    def createTreeWidget(self):
        """ createTreeWidget() -> QModuleTreeWidget
        Return the search tree widget for this window

        """
        self.setWindowTitle('Modules')
        return QModuleTreeWidget(self)

    def findModule(self, descriptor):
        moduleName = descriptor.name

        items = [x for x in
                 self.treeWidget.findItems(moduleName,
                                           QtCore.Qt.MatchExactly |
                                           QtCore.Qt.MatchWrap |
                                           QtCore.Qt.MatchRecursive)
                 if not x.is_top_level() and x.descriptor == descriptor]
        if len(items) <> 1:
            raise VistrailsInternalError("Expected one item (%s), got %d: %s" %
                                         (moduleName,
                                          len(items),
                                          ";".join(x.descriptor.name
                                                   for x in items)))
        item = items[0]
        return item

    def showModule(self, descriptor):
        item = self.findModule(descriptor)
        item.setHidden(False)

    def hideModule(self, descriptor):
        item = self.findModule(descriptor)
        item.setHidden(True)

    def switchDescriptors(self, old_descriptor, new_descriptor):
        old_item = self.findModule(old_descriptor)
        new_item = self.findModule(new_descriptor)
        old_item.set_descriptor(new_descriptor)
        new_item.set_descriptor(old_descriptor)

    def deletedModule(self, descriptor):
        """ deletedModule(descriptor: core.modules.module_registry.ModuleDescriptor)
        A module has been removed from VisTrails

        """
        if descriptor.module_abstract():
            # skip abstract and hidden modules, they're not in the tree
            return

        item = self.findModule(descriptor)
        parent = item.parent()
        parent.takeChild(parent.indexOfChild(item))
        del item.descriptor

    def deletedPackage(self, package):
        """ deletedPackage(package):
        A package has been deleted from VisTrails
        """
        try:
            item = self.packages[package.identifier]
        except KeyError:
            pass
        else:
            index = self.treeWidget.indexOfTopLevelItem(item)
            self.treeWidget.takeTopLevelItem(index)
            del self.packages[package.identifier]

    def newPackage(self, package_identifier, prepend=False):
        # prepend places at the front of the list of packages,
        # by default adds to the end of the list of packages
        # Right now the list is sorted so prepend has no effect
        if package_identifier in self.packages:
            return self.packages[package_identifier]
        registry = get_module_registry()
        package_name = registry.packages[package_identifier].name
        package_item = QPackageTreeWidgetItem(None,
                                              package_name, package_identifier)
        self.packages[package_identifier] = package_item
        if prepend:
            self.treeWidget.insertTopLevelItem(0, package_item)
        else:
            self.treeWidget.addTopLevelItem(package_item)
        return package_item

    def newModule(self, descriptor, recurse=False):
        """ newModule(descriptor: core.modules.module_registry.ModuleDescriptor)
        A new module has been added to VisTrails

        """
        if not descriptor.module_abstract(): # and not descriptor.is_hidden:
            # skip abstract modules, they're no longer in the tree

            # NB: only looks for toplevel matches
            package_identifier = descriptor.ghost_identifier or descriptor.identifier
            if package_identifier not in self.packages:
                package_item = self.newPackage(package_identifier, True)
            else:
                package_item = self.packages[package_identifier]

            if descriptor.ghost_namespace is not None:
                namespace = descriptor.ghost_namespace
            else:
                namespace = descriptor.namespace
            if descriptor.namespace_hidden or not namespace:
                parent_item = package_item
            else:
                parent_item = \
                    package_item.get_namespace(namespace.split('|'))

            item = QModuleTreeWidgetItem(descriptor, parent_item,
                                         [descriptor.name],
                                         descriptor.is_hidden)
            if descriptor.is_hidden:
                item.setHidden(True)
        if recurse:
            for child in descriptor.children:
                self.newModule(child, recurse)

    def updateFromModuleRegistry(self):
        """ updateFromModuleRegistry(packageName: str) -> None
        Setup this tree widget to show modules currently inside
        module_registry.registry

        """

        self.treeWidget.setSortingEnabled(False)
        registry = get_module_registry()
        for package in registry.package_list:
            self.newPackage(package.identifier)
        self.newModule(registry.root_descriptor, True)
        self.treeWidget.sortByColumn(0, QtCore.Qt.AscendingOrder)
        self.treeWidget.setSortingEnabled(True)

    def sizeHint(self):
        return QtCore.QSize(256, 760)

class QModuleTreeWidget(QSearchTreeWidget):
    """
    QModuleTreeWidget is a subclass of QSearchTreeWidget to display all
    Vistrails Module

    """
    def __init__(self, parent=None):
        """ QModuleTreeWidget(parent: QWidget) -> QModuleTreeWidget
        Set up size policy and header

        """
        QSearchTreeWidget.__init__(self, parent)
        self.header().hide()
        self.setRootIsDecorated(False)
        self.delegate = QModuleTreeWidgetItemDelegate(self, self)
        self.setItemDelegate(self.delegate)
        self.connect(self,
                     QtCore.SIGNAL('itemPressed(QTreeWidgetItem *,int)'),
                     self.onItemPressed)

    def onItemPressed(self, item, column):
        """ onItemPressed(item: QTreeWidgetItem, column: int) -> None
        Expand/Collapse top-level item when the mouse is pressed

        """
        if item and item.parent() is None:
            self.setItemExpanded(item, not self.isItemExpanded(item))

    def contextMenuEvent(self, event):
        # Just dispatches the menu event to the widget item
        item = self.itemAt(event.pos())
        if item:
            # find top level
            p = item
            while p.parent():
                p = p.parent()
            # get package identifier
            assert isinstance(p, QPackageTreeWidgetItem)
            identifier = p.identifier
            registry = get_module_registry()
            package = registry.packages[identifier]
            try:
                if package.has_contextMenuName():
                    name = package.contextMenuName(item.text(0))
                    if name:
                        act = QtGui.QAction(name, self)
                        act.setStatusTip(name)
                        def callMenu():
                            if package.has_callContextMenu():
                                name = package.callContextMenu(item.text(0))

                        QtCore.QObject.connect(act,
                                               QtCore.SIGNAL("triggered()"),
                                               callMenu)
                        menu = QtGui.QMenu(self)
                        menu.addAction(act)
                        menu.exec_(event.globalPos())
                    return
            except Exception, e:
                debug.warning("Got exception trying to display %s's "
                              "context menu in the palette: %s: %s" % (
                              package.name,
                              type(e).__name__, ', '.join(e.args)))

            item.contextMenuEvent(event, self)

    def startDrag(self, actions):
        indexes = self.selectedIndexes()
        if len(indexes) > 0:
            mime_data = self.model().mimeData(indexes)
            drag = QtGui.QDrag(self)
            drag.setMimeData(mime_data)
            item = mime_data.items[0]

            app = get_vistrails_application()
            pipeline_view = app.builderWindow.get_current_controller().current_pipeline_view
            if hasattr(pipeline_view.scene(), 'add_tmp_module'):
                module_item = \
                    pipeline_view.scene().add_tmp_module(item.descriptor)
                pixmap = pipeline_view.paintModuleToPixmap(module_item)

                drag.setPixmap(pixmap)
                drag.setHotSpot(QtCore.QPoint(pixmap.width()/2,
                                              pixmap.height()/2))
                drag.exec_(actions)
                pipeline_view.scene().delete_tmp_module()


class QModuleTreeWidgetItemDelegate(QtGui.QItemDelegate):
    """
    QModuleTreeWidgetItemDelegate will override the original
    QTreeWidget paint function to draw buttons for top-level item
    similar to QtDesigner. This mimics
    Qt/tools/designer/src/lib/shared/sheet_delegate, which is only a
    private class from QtDesigned.

    """
    def __init__(self, view, parent):
        """ QModuleTreeWidgetItemDelegate(view: QTreeView,
                                          parent: QWidget)
                                          -> QModuleTreeWidgetItemDelegate
        Create the item delegate given the tree view

        """
        QtGui.QItemDelegate.__init__(self, parent)
        self.treeView = view
        self.isMac = systemType == 'Darwin'

    def paint(self, painter, option, index):
        """ painter(painter: QPainter, option QStyleOptionViewItem,
                    index: QModelIndex) -> None
        Repaint the top-level item to have a button-look style

        """
        model = index.model()
        if not model.parent(index).isValid():
            buttonOption = QtGui.QStyleOptionButton()
            buttonOption.state = option.state
            if self.isMac:
                buttonOption.state |= QtGui.QStyle.State_Raised
            buttonOption.state &= ~QtGui.QStyle.State_HasFocus

            buttonOption.rect = option.rect
            buttonOption.palette = option.palette
            buttonOption.features = QtGui.QStyleOptionButton.None

            style = self.treeView.style()

            style.drawControl(QtGui.QStyle.CE_PushButton,
                              buttonOption,
                              painter,
                              self.treeView)

            branchOption = QtGui.QStyleOption()
            i = 9 ### hardcoded in qcommonstyle.cpp
            r = option.rect
            branchOption.rect = QtCore.QRect(r.left() + i / 2,
                                             r.top() + (r.height() - i) / 2,
                                             i, i)
            branchOption.palette = option.palette
            branchOption.state = QtGui.QStyle.State_Children

            if self.treeView.isExpanded(index):
                branchOption.state |= QtGui.QStyle.State_Open

            style.drawPrimitive(QtGui.QStyle.PE_IndicatorBranch,
                                branchOption,
                                painter, self.treeView)

            textrect = QtCore.QRect(r.left() + i * 2,
                                    r.top(),
                                    r.width() - ((5 * i) / 2),
                                    r.height())
            text = option.fontMetrics.elidedText(
                model.data(index,
                           QtCore.Qt.DisplayRole),
                QtCore.Qt.ElideMiddle,
                textrect.width())
            style.drawItemText(painter,
                               textrect,
                               QtCore.Qt.AlignCenter,
                               option.palette,
                               self.treeView.isEnabled(),
                               text)
        else:
            QtGui.QItemDelegate.paint(self, painter, option, index)

    def sizeHint(self, option, index):
        """ sizeHint(option: QStyleOptionViewItem, index: QModelIndex) -> None
        Take into account the size of the top-level button

        """
        return (QtGui.QItemDelegate.sizeHint(self, option, index) +
                QtCore.QSize(2, 2))



class QModuleTreeWidgetItem(QtGui.QTreeWidgetItem):
    """
    QModuleTreeWidgetItem represents module on QModuleTreeWidget

    """

    def __init__(self, descriptor, parent, labelList, is_hidden):
        """ QModuleTreeWidgetItem(descriptor: ModuleDescriptor
                                    (or None for top-level),
                                  parent: QTreeWidgetItem
                                  labelList: string)
                                  -> QModuleTreeWidget
        Create a new tree widget item with a specific parent and
        labels

        """
        QtGui.QTreeWidgetItem.__init__(self, parent, labelList)
        self.set_descriptor(descriptor)

        # Real flags store the widget's flags prior to masking related
        # to abstract modules, etc.
        self._real_flags = self.flags()

        # This is necessary since we override setFlags
        self.setFlags(self._real_flags)

        self.is_hidden = is_hidden

    def added_input_port(self):
        self.setFlags(self._real_flags)

    def added_output_port(self):
        self.setFlags(self._real_flags)

    def setFlags(self, flags):
        self._real_flags = flags
        d = self.descriptor
        if d is None:
            # toplevel moduletree widgets are never draggable
            flags = flags & ~QtCore.Qt.ItemIsDragEnabled
        elif d.module_abstract():
            # moduletree widgets for abstract modules are never
            # draggable or enabled
            flags = flags & ~(QtCore.Qt.ItemIsDragEnabled |
                              QtCore.Qt.ItemIsSelectable)
        QtGui.QTreeWidgetItem.setFlags(self, flags)

    def is_top_level(self):
        return self.descriptor is None

    def contextMenuEvent(self, event, widget):
        if self.is_top_level():
            return
        menu = QtGui.QMenu(widget)
        act = QtGui.QAction("View Documentation", widget)
        act.setStatusTip("View module documentation")
        QtCore.QObject.connect(act,
                               QtCore.SIGNAL("triggered()"),
                               self.view_documentation)
        menu.addAction(act)
        if self.descriptor.package == 'local.abstractions':
            act = QtGui.QAction("Edit Subworkflow", widget)
            act.setStatusTip("Edit this Subworkflow")
            QtCore.QObject.connect(act,
                               QtCore.SIGNAL("triggered()"),
                               self.edit_subworkflow)
            menu.addAction(act)
        menu.exec_(event.globalPos())

    def view_documentation(self):
        from vistrails_window import _app
        _app.show_documentation()
        widget = QModuleDocumentation.instance()
        widget.update_descriptor(self.descriptor)

    def edit_subworkflow(self):
        from vistrails_window import _app
        filename = self.descriptor.module.vt_fname
        _app.openAbstraction(filename)

    def set_descriptor(self, descriptor):
        self.descriptor = descriptor
        if descriptor:
            descriptor.set_widget(self)

class QNamespaceTreeWidgetItem(QModuleTreeWidgetItem):
    def __init__(self, parent, name):
        QModuleTreeWidgetItem.__init__(self, None, parent, [name], False)
        self.setFlags(self.flags() & ~QtCore.Qt.ItemIsDragEnabled)
        self.namespaces = {}

    def get_namespace(self, namespace_items):
        if len(namespace_items) <= 0:
            return self
        namespace_item = namespace_items.pop(0)
        if namespace_item in self.namespaces and self.namespaces[namespace_item]:
            item = self.namespaces[namespace_item]
        else:
            item = QNamespaceTreeWidgetItem(self, namespace_item)
            self.namespaces[namespace_item] = item
        return item.get_namespace(namespace_items)

    def takeChild(self, index):
        child = self.child(index)
        if child.text(0) in self.namespaces:
            del self.namespaces[child.text(0)]
        QModuleTreeWidgetItem.takeChild(self, index)
        if self.childCount() < 1 and self.parent():
            self.parent().takeChild(self.parent().indexOfChild(self))

class QPackageTreeWidgetItem(QNamespaceTreeWidgetItem):
    def __init__(self, parent, name, identifier):
        QNamespaceTreeWidgetItem.__init__(self, parent, name)
        self.identifier = identifier



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import ast
from xml.etree import cElementTree as ET

class SpecList(object):
    """ A class with module specifications and custom code
        This describes how the wrapped methods/classes will
        maps to modules in vistrails
    """

    def __init__(self, module_specs=None):
        if module_specs is None:
            module_specs = []
        self.module_specs = module_specs

    def write_to_xml(self, fname):
        root = ET.Element("specs")
        for spec in self.module_specs:
            root.append(spec.to_xml())
        tree = ET.ElementTree(root)

        def indent(elem, level=0):
            i = "\n" + level*"  "
            if len(elem):
                if not elem.text or not elem.text.strip():
                    elem.text = i + "  "
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
                for elem in elem:
                    indent(elem, level+1)
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
            else:
                if level and (not elem.tail or not elem.tail.strip()):
                    elem.tail = i
        indent(tree.getroot())

        tree.write(fname)

    @staticmethod
    def read_from_xml(fname, klass=None):
        if klass is None:
            klass = ModuleSpec
        module_specs = []
        tree = ET.parse(fname)
        for elt in tree.getroot():
            if elt.tag == klass.xml_name:
                module_specs.append(klass.from_xml(elt))
        retval = SpecList(module_specs)
        # for spec in retval.module_specs:
        #     print "==", spec.name, "=="
        #     for ps in spec.port_specs:
        #         print " ", ps.arg, ps.name
        return retval

######### BASE MODULE SPEC ###########

class PortSpec(object):
    """ Represents specification of a port
    """
    xml_name = "portSpec"
    # attrs tuple means (default value, [is subelement, [run eval]])
    # Subelement: ?
    # eval: serialize as string and use eval to get value back
    # FIXME: subelement/eval not needed if using json
    attrs = {"name": "",                         # port name
             "port_type": None,                  # type signature in vistrails
             "docstring": ("", True),            # documentation
             "min_conns": (0, False, True),      # set min_conns (1=required)
             "max_conns": (-1, False, True),     # Set max_conns (default -1)
             "show_port": (False, False, True),  # Set not optional (use connection)
             "sort_key": (-1, False, True),      # sort_key
             "shape": (None, False, True),       # physical shape
             "depth": (0, False, True)}          # expected list depth

    def __init__(self, **kwargs):
        self.set_defaults(**kwargs)
        self.port_types = []

    def set_defaults(self, **kwargs):
        for attr, props in self.attrs.iteritems():
            if isinstance(props, tuple):
                default_val = props[0]
            else:
                default_val = props
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])
            else:
                setattr(self, attr, default_val)

    def to_xml(self, elt=None):
        if elt is None:
            elt = ET.Element(self.xml_name)
        for attr, props in self.attrs.iteritems():
            attr_val = getattr(self, attr)
            is_subelt = False
            if isinstance(props, tuple):
                default_val = props[0]
                if len(props) > 1:
                    is_subelt = props[1]
            else:
                default_val = props

            if default_val != attr_val:
                if is_subelt:
                    subelt = ET.Element(attr)
                    subelt.text = unicode(getattr(self, attr))
                    elt.append(subelt)
                else:
                    elt.set(attr, unicode(attr_val))
        return elt

    @classmethod
    def internal_from_xml(cls, elt, obj=None):
        if obj is None:
            obj = cls()

        child_elts = {}
        for child in elt.getchildren():
            # if child.tag not in obj.attrs:
            #     raise RuntimeError('Cannot deal with tag "%s"' % child.tag)
            if child.tag not in child_elts:
                child_elts[child.tag] = []
            child_elts[child.tag].append(child)

        kwargs = {}
        for attr, props in obj.attrs.iteritems():
            is_subelt = False
            run_eval = False
            if isinstance(props, tuple):
                if len(props) > 1:
                    is_subelt = props[1]
                if len(props) > 2:
                    run_eval = props[2]
            attr_vals = []
            if is_subelt:
                if attr in child_elts:
                    attr_vals = [c.text for c in child_elts[attr]
                                 if c.text is not None]
            else:
                attr_val = elt.get(attr)
                if attr_val is not None:
                    attr_vals = [attr_val]

            if len(attr_vals) > 1:
                raise ValueError('Should have only one value for '
                                'attribute "%s"' % attr)
            if len(attr_vals) > 0:
                attr_val = attr_vals[0]
                if run_eval:
                    try:
                        kwargs[attr] = ast.literal_eval(attr_val)
                    except (NameError, SyntaxError, ValueError):
                        kwargs[attr] = attr_val
                else:
                    kwargs[attr] = attr_val
        obj.set_defaults(**kwargs)
        return obj, child_elts

    @classmethod
    def from_xml(cls, elt, obj=None):
        obj, child_elts = cls.internal_from_xml(elt, obj)
        return obj

    @classmethod
    def create_from_xml(cls, elt):
        if elt.tag == cls.InputSpecType.xml_name:
            return cls.InputSpecType.from_xml(elt)
        elif elt.tag == cls.OutputSpecType.xml_name:
            return cls.OutputSpecType.from_xml(elt)
        raise TypeError('Cannot create spec from element of type "%s"' %
                        elt.tag)

    def get_port_type(self):
        if self.port_type is None:
            return "basic:Null"
        try:
            port_types = ast.literal_eval(self.port_type)
            def flatten(t):
                if not isinstance(t, list):
                    raise Exception("Expected a list")
                flat = []
                for elt in t:
                    if isinstance(elt, list):
                        flat.extend(flatten(elt))
                    else:
                        flat.append(elt)
                return flat
            return ','.join(flatten(port_types))
        except (SyntaxError, ValueError):
            pass
        return self.port_type

    def get_prepend_params(self):
        if self.prepend_params is None:
            return []
        return self.prepend_params


class InputPortSpec(PortSpec):
    xml_name = "inputPortSpec"
    attrs = {"entry_types": (None, True, True),# custom entry type (like enum)
             "values": (None, True, True),     # values for enums
             "labels": (None, True, True),   # custom labels on enum values
             "defaults": (None, True, True),   # default value list
             }
    attrs.update(PortSpec.attrs)

    def get_port_attrs(self):
        """ Port attribute dict that will be used to create the port

        """
        attrs = {}
        if self.sort_key != -1:
            attrs["sort_key"] = self.sort_key
        if self.shape:
            attrs["shape"] = self.shape
        if self.depth:
            attrs["depth"] = self.depth
        if self.values:
            attrs["values"] = unicode(self.values)
        if self.labels:
            attrs["labels"] = unicode(self.labels)
        if self.entry_types:
            attrs["entry_types"] = unicode(self.entry_types)
        if self.defaults:
            attrs["defaults"] = unicode(self.defaults)
        if self.docstring:
            attrs["docstring"] = self.docstring
        if self.min_conns:
            attrs["min_conns"] = self.min_conns
        if self.max_conns != -1:
            attrs["max_conns"] = self.max_conns
        if not self.show_port:
            attrs["optional"] = True
        return attrs


class OutputPortSpec(PortSpec):
    xml_name = "outputPortSpec"
    attrs = {}
    attrs.update(PortSpec.attrs)

    def get_port_attrs(self):
        """ Port attribute dict that will be used to create the port

        """
        attrs = {}
        if self.sort_key != -1:
            attrs["sort_key"] = self.sort_key
        if self.shape:
            attrs["shape"] = self.shape
        if self.depth:
            attrs["depth"] = self.depth
        if self.docstring:
            attrs["docstring"] = self.docstring
        if self.min_conns:
            attrs["min_conns"] = self.min_conns
        if self.max_conns != -1:
            attrs["max_conns"] = self.max_conns
        if not self.show_port:
            attrs["optional"] = True
        return attrs

class ModuleSpec(object):
    """ Represents specification of a module
        This mirrors how the module will look in the vistrails registry
    """
    xml_name = 'moduleSpec'
    InputSpecType = InputPortSpec
    OutputSpecType = OutputPortSpec

    # From Modulesettings. See core.modules.config._documentation
    ms_attrs = ['name',
                'configure_widget',
                'constant_widget',
                'constant_widgets',
                'signature',
                'constant_signature',
                'color',
                'fringe',
                'left_fringe',
                'right_fringe',
                'abstract',
                'namespace',
                'package_version',
                'hide_descriptor']
    attrs = [
             # basic attributes
             'module_name', # Name of module (can be overridden by modulesettings)
             'superklass',  # class to inherit from
             'code_ref',    # reference to wrapped class/method
             'docstring',   # module __doc__
             'cacheable',   # should this module be cached
             # special attributes
             'callback',    # name of attribute for progress callback
             'tempfile']    # attribute name for temporary file creation method
    attrs.extend(ms_attrs)

    def __init__(self, module_name='', superklass='', code_ref='',
                 docstring='', callback=None, tempfile=None, cacheable=True,
                 input_port_specs=None, output_port_specs=None, **kwargs):
        if input_port_specs is None:
            input_port_specs = []
        if output_port_specs is None:
            output_port_specs = []

        self.module_name = module_name
        self.superklass = superklass
        self.code_ref = code_ref
        self.docstring = docstring
        self.callback = callback
        self.tempfile = tempfile
        self.cacheable = cacheable

        self.input_port_specs = input_port_specs
        self.output_port_specs = output_port_specs

        for attr in self.ms_attrs:
            setattr(self, attr, kwargs.get(attr, None))

    def to_xml(self, elt=None):
        if elt is None:
            elt = ET.Element(self.xml_name)
        elt.set("module_name", self.module_name)
        elt.set("superklass", self.superklass)
        elt.set("code_ref", self.code_ref)
        subelt = ET.Element("docstring")
        subelt.text = unicode(self.docstring)
        elt.append(subelt)
        if self.callback is not None:
            elt.set("callback", self.callback)
        if self.tempfile is not None:
            elt.set("tempfile", self.tempfile)
        if self.cacheable is False:
            elt.set("cacheable", 'False')
        for attr in self.ms_attrs:
            value = getattr(self, attr)
            if value is not None:
                elt.set(attr, repr(value))

        for port_spec in self.input_port_specs:
            subelt = port_spec.to_xml()
            elt.append(subelt)
        for port_spec in self.output_port_specs:
            subelt = port_spec.to_xml()
            elt.append(subelt)
        return elt

    @staticmethod
    def from_xml(elt, klass=None):
        if klass is None:
            klass = ModuleSpec
        module_name = elt.get("module_name", '')
        superklass = elt.get("superklass", '')
        code_ref = elt.get("code_ref", '')
        callback = elt.get("callback", None)
        tempfile = elt.get("tempfile", None)
        cacheable = ast.literal_eval(elt.get("cacheable", "True"))

        kwargs = {}
        for attr in klass.ms_attrs:
            value = elt.get(attr, None)
            if value is not None:
                kwargs[attr] = ast.literal_eval(value)

        docstring = ""
        input_port_specs = []
        output_port_specs = []
        for child in elt.getchildren():
            if child.tag == klass.InputSpecType.xml_name:
                input_port_specs.append(klass.InputSpecType.from_xml(child))
            elif child.tag == klass.OutputSpecType.xml_name:
                output_port_specs.append(klass.OutputSpecType.from_xml(child))
            elif child.tag == "docstring":
                if child.text:
                    docstring = child.text
        return klass(module_name=module_name, superklass=superklass,
                   code_ref=code_ref, docstring=docstring,
                   callback=callback, tempfile=tempfile, cacheable=cacheable,
                   input_port_specs=input_port_specs,
                   output_port_specs=output_port_specs, **kwargs)

    def get_output_port_spec(self, compute_name):
        for ps in self.output_port_specs:
            if ps.compute_name == compute_name:
                return ps
        return None

    def get_module_settings(self):
        """ Returns modulesettings dict

        """
        attrs = {}
        for attr in self.ms_attrs:
            value = getattr(self, attr)
            if value is not None:
                attrs[attr] = value
        return attrs


######### PYTHON FUNCTION SPEC ###########

class FunctionInputPortSpec(InputPortSpec):
    xml_name = "functionInputPortSpec"
    attrs = {"arg": ""}                  # attribute name
    attrs.update(InputPortSpec.attrs)


class FunctionOutputPortSpec(OutputPortSpec):
    xml_name = "functionOutputPortSpec"


class FunctionSpec(ModuleSpec):
    """ Specification for wrapping a python function
    """
    xml_name = 'functionSpec'
    InputSpecType = FunctionInputPortSpec
    OutputSpecType = FunctionOutputPortSpec

    attrs = ['output_type'] # None(=single), list(ordered), or dict(attr=value)
    attrs.extend(ModuleSpec.attrs)

    def __init__(self, module_name, superklass='', code_ref='', docstring="",
                 output_type=None, callback=None, tempfile=None,
                 cacheable=True, input_port_specs=None, output_port_specs=None,
                 **kwargs):
        ModuleSpec.__init__(self, module_name, superklass, code_ref,
                            docstring, callback, tempfile, cacheable,
                            input_port_specs, output_port_specs, **kwargs)
        self.output_type = output_type

    def to_xml(self, elt=None):
        if elt is None:
            elt = ET.Element(self.xml_name)
        elt = ModuleSpec.to_xml(self, elt)
        if self.output_type is not None:
            elt.set("output_type", self.output_type)
        return elt

    @staticmethod
    def from_xml(elt):
        inst = ModuleSpec.from_xml(elt, FunctionSpec)
        inst.output_type = elt.get("output_type", None)
        return inst

######### PYTHON CLASS SPEC ###########


class ClassInputPortSpec(InputPortSpec):
    xml_name = "classInputPortSpec"
    attrs = {"method_name": "",                  # method name
             "method_type": "",                  # Type like nullary, OnOff or SetXToY
             "prepend_params": (None, True, True)} # prepended params like index
    attrs.update(InputPortSpec.attrs)

    def __init__(self, **kwargs):
        InputPortSpec.__init__(self, **kwargs)
        if not self.method_name:
            self.method_name = self.name


class ClassOutputPortSpec(OutputPortSpec):
    xml_name = "classOutputPortSpec"
    attrs = {"method_name": "",                    # method/attribute name
             "prepend_params": (None, True, True)} # prepended params used with indexed methods
    attrs.update(OutputPortSpec.attrs)

    def __init__(self, **kwargs):
        OutputPortSpec.__init__(self, **kwargs)
        if not self.method_name:
            self.method_name = self.name


class ClassSpec(ModuleSpec):
    """ Specification for wrapping a python class
    """
    xml_name = 'classSpec'
    InputSpecType = ClassInputPortSpec
    OutputSpecType = ClassOutputPortSpec
    attrs = ['methods_last', # If True will compute methods before connections
             'compute', # Function to call after input methods
             'cleanup'] # Function to call after output methods
    attrs.extend(ModuleSpec.attrs)

    def __init__(self, module_name, superklass='', code_ref='', docstring="",
                 callback=None, tempfile=None,
                 cacheable=True, input_port_specs=None, output_port_specs=None,
                 compute=None, cleanup=None, methods_last=False, **kwargs):
        ModuleSpec.__init__(self, module_name, superklass, code_ref,
                            docstring, callback, tempfile, cacheable,
                            input_port_specs, output_port_specs, **kwargs)
        self.methods_last = methods_last
        self.compute = compute
        self.cleanup = cleanup

    def to_xml(self, elt=None):
        if elt is None:
            elt = ET.Element(self.xml_name)
        if self.methods_last is not False:
            elt.set("methods_last", unicode(self.methods_last))
        if self.compute is not None:
            elt.set("compute", self.compute)
        if self.cleanup is not None:
            elt.set("cleanup", self.cleanup)
        elt = ModuleSpec.to_xml(self, elt)
        return elt

    @staticmethod
    def from_xml(elt):
        inst = ModuleSpec.from_xml(elt, ClassSpec)
        inst.methods_last = ast.literal_eval(elt.get("methods_last", 'False'))
        inst.compute = elt.get("compute", None)
        inst.cleanup = elt.get("cleanup", None)
        return inst

###############################################################################

import unittest


class TestModuleSpec(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            import vtk
        except ImportError:
            raise unittest.SkipTest("vtk is not installed")

        from vistrails.tests.utils import enable_package
        from ..identifiers import identifier

        enable_package(identifier)

    def test_module_spec(self):
        input_spec = InputPortSpec(name='myportname',
                                   port_type='basic:String',
                                   docstring='my port doc',
                                   min_conns=1,
                                   max_conns=3,
                                   show_port=True,
                                   sort_key=5,
                                   depth=1,
                                   entry_type='enum')
        in_attrs = input_spec.get_port_attrs()

        output_spec = OutputPortSpec(name='myportname',
                                   port_type='basic:String',
                                   docstring='my port doc',
                                   min_conns=1,
                                   max_conns=3,
                                   show_port=False,
                                   sort_key=5,
                                   depth=1)
        out_attrs = output_spec.get_port_attrs()

        ms = ModuleSpec(module_name='myclassname',
                        superklass='mysuperclassname',
                        code_ref='theclassname',
                        docstring='my documentation',
                        callback=None,
                        tempfile=None,
                        cacheable=False,
                        input_port_specs=[input_spec],
                        output_port_specs=[output_spec])
        as_string = ET.tostring(ms.to_xml())
        from_string = ET.fromstring(as_string)
        ms2 = ModuleSpec.from_xml(from_string)
        in_attrs2 = ms2.input_port_specs[0].get_port_attrs()
        out_attrs2 = ms2.output_port_specs[0].get_port_attrs()
        self.assertEqual(in_attrs, in_attrs2)
        self.assertEqual(out_attrs, out_attrs2)

    def test_function_spec(self):
        input_spec = FunctionInputPortSpec(name='myportname',
                                   port_type='basic:String',
                                   docstring='my port doc',
                                   min_conns=1,
                                   max_conns=3,
                                   show_port=False,
                                   sort_key=5,
                                   depth=1,
                                   arg='myargname',
                                   )
        in_attrs = input_spec.get_port_attrs()

        output_spec = FunctionOutputPortSpec(name='myportname',
                                   port_type='basic:String',
                                   docstring='my port doc',
                                   min_conns=1,
                                   max_conns=3,
                                   show_port=False,
                                   sort_key=5,
                                   depth=1)
        out_attrs = output_spec.get_port_attrs()

        ms = FunctionSpec(module_name='myclassname',
                        superklass='mysuperclassname',
                        code_ref='theclassname',
                        docstring='my documentation',
                        callback=None,
                        tempfile=None,
                        cacheable=False,
                        input_port_specs=[input_spec],
                        output_port_specs=[output_spec],
                        output_type='list')
        as_string = ET.tostring(ms.to_xml())
        from_string = ET.fromstring(as_string)
        ms2 = FunctionSpec.from_xml(from_string)
        in_attrs2 = ms2.input_port_specs[0].get_port_attrs()
        out_attrs2 = ms2.output_port_specs[0].get_port_attrs()
        self.assertEqual(in_attrs, in_attrs2)
        self.assertEqual(out_attrs, out_attrs2)

    def test_class_spec(self):
        input_spec = ClassInputPortSpec(name='myportname',
                                   port_type='basic:String',
                                   docstring='my port doc',
                                   min_conns=1,
                                   max_conns=3,
                                   show_port=False,
                                   sort_key=5,
                                   depth=1,
                                   method_name='MyClassMethodName',
                                   method_type='SetXToY',
                                   prepend_params=[1])
        in_attrs = input_spec.get_port_attrs()

        output_spec = ClassOutputPortSpec(name='myportname',
                                   port_type='basic:String',
                                   docstring='my port doc',
                                   min_conns=1,
                                   max_conns=3,
                                   show_port=False,
                                   sort_key=5,
                                   depth=1,
                                   method_name='MyClassMethodName',
                                   prepend_params=[1])
        out_attrs = output_spec.get_port_attrs()

        ms = ClassSpec(module_name='myclassname',
                        superklass='mysuperclassname',
                        code_ref='theclassname',
                        docstring='my documentation',
                        callback=None,
                        tempfile=None,
                        cacheable=False,
                        input_port_specs=[input_spec],
                        output_port_specs=[output_spec],
                        methods_last=True,
                        compute='myCompute',
                        cleanup='myCleanup')
        as_string = ET.tostring(ms.to_xml())
        from_string = ET.fromstring(as_string)
        ms2 = ClassSpec.from_xml(from_string)
        in_attrs2 = ms2.input_port_specs[0].get_port_attrs()
        out_attrs2 = ms2.output_port_specs[0].get_port_attrs()
        self.assertEqual(in_attrs, in_attrs2)
        self.assertEqual(out_attrs, out_attrs2)


#def run():
#    specs = SpecList.read_from_xml("mpl_plots_raw.xml")
#    specs.write_to_xml("mpl_plots_raw_out.xml")

#if __name__ == '__main__':
#    run()



"""Table data package for VisTrails.

This package contains data extraction and manipulation facilities. It wraps
numpy and csv and allows the use of several file types from VisTrails, with
extraction and conversion routines.

"""

# ChangeLog:
# 2014-05-29 -- 0.1.5
#   * Updates JSON readers
# 2014-02-03 -- 0.1.4
#   * Merges some tabdata functionality (improve CSV)
# 2014-01-14 -- 0.1.3
#   * Adds writer modules
# 2013-12-09 -- 0.1.2
#   * No longer use 'self' ports to output tables
# 2013-10-04 -- 0.1.1
#   * Moves reading modules out of specific namespaces
#     (read|csv|CSVFile -> read|CSVFile)
# 2013-05-16 -- 0.1.0
#   * Package created (for DAT project)

from __future__ import division

from vistrails.core.packagemanager import get_package_manager

from .identifiers import *


def package_dependencies():
    pm = get_package_manager()
    spreadsheet_identifier = 'org.vistrails.vistrails.spreadsheet'
    if pm.has_package(spreadsheet_identifier):
        return [spreadsheet_identifier]
    else: # pragma: no cover
        return []


def package_requirements():
    from vistrails.core.requirements import require_python_module
    require_python_module('csv')



#!/usr/bin/env python
# pragma: no testimport
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""Main file for the VisTrails distribution."""

from __future__ import division

import os
import sys

# Allows the userpackages directory to be overridden through an environment
# variable
# As this variable is set by the package manager, this also allows
# multiprocessing to work correctly on Windows, where fork is not used and thus
# 'userpackages' needs to be available in processes spawned from a
# userpackage's code
try:
    userpackages_dir = os.environ['VISTRAILS_USERPACKAGES_DIR']
except KeyError:
    pass
else:
    old_sys_path = list(sys.path)
    sys.path.insert(0, os.path.join(userpackages_dir, os.path.pardir))
    try:
        import userpackages
    except ImportError:
        sys.stderr.write("Couldn't import VISTRAILS_USERPACKAGES_DIR (%s), "
                         "continuing\n" % userpackages_dir)

def disable_lion_restore():
    """ Prevent Mac OS 10.7 to restore windows state since it would
    make Qt 4.7.3 unstable due to its lack of handling Cocoa's Main
    Window. """
    import platform
    if platform.system()!='Darwin': return
    release = platform.mac_ver()[0].split('.')
    if len(release)<2: return
    major = int(release[0])
    minor = int(release[1])
    if major*100+minor<107: return
    ssPath = os.path.expanduser('~/Library/Saved Application State/org.vistrails.savedState')
    if os.path.exists(ssPath):
        os.system('rm -rf "%s"' % ssPath)
    os.system('defaults write org.vistrails NSQuitAlwaysKeepsWindows -bool false')


def fix_site():
    # py2app ships a stripped version of site.py
    # USER_BASE and USER_SITE is not set,
    # this is needed by at least scipy.weave
    import platform
    if platform.system()!='Darwin': return
    import site
    if hasattr(site, "USER_BASE"): return
    from vistrails.core.system import mac_site
    site.USER_BASE = mac_site.getuserbase()
    site.USER_SITE = mac_site.getusersitepackages()
    site._Helper = mac_site._Helper

def fix_paths():
    import site
    if not hasattr(site, "USER_BASE"): return # We are running py2app

    # Fix import path: add parent directory(so that we can
    # import vistrails.[gui|...] and remove other paths below it (we might have
    # been started from a subdir)
    # A better solution is probably to move run.py up a
    # directory in the repo
    old_dir = os.path.realpath(os.path.dirname(__file__))
    vistrails_dir = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))
    i = 0
    while i < len(sys.path):
        rpath = os.path.realpath(sys.path[i])
        if rpath.startswith(old_dir):
            del sys.path[i]
        else:
            i += 1
    if vistrails_dir not in sys.path:
        sys.path.insert(0, vistrails_dir)

def main():
    fix_paths()
    disable_lion_restore()
    fix_site()

    # Load the default locale (from environment)
    import locale
    locale.setlocale(locale.LC_ALL, '')

    # Log to the console
    from vistrails.core import debug
    debug.DebugPrint.getInstance().log_to_console()

    from vistrails.gui.requirements import require_pyqt4_api2
    require_pyqt4_api2()

    import vistrails.gui.application
    from vistrails.core.application import APP_SUCCESS, APP_FAIL, APP_DONE
    try:
        v = vistrails.gui.application.start_application(args=sys.argv[1:])
        if v != APP_SUCCESS:
            app = vistrails.gui.application.get_vistrails_application()
            if app:
                app.finishSession()
            sys.exit(APP_SUCCESS if v == APP_DONE else APP_FAIL)
        app = vistrails.gui.application.get_vistrails_application()()
    except SystemExit, e:
        app = vistrails.gui.application.get_vistrails_application()
        if app:
            app.finishSession()
        sys.exit(e)
    except Exception, e:
        app = vistrails.gui.application.get_vistrails_application()
        if app:
            app.finishSession()
        import traceback
        print >>sys.stderr, "Uncaught exception on initialization: %s" % (
                traceback._format_final_exc_line(type(e).__name__, e).strip())
        traceback.print_exc(None, sys.stderr)
        sys.exit(255)
    if (not app.temp_configuration.batch and
        not app.temp_configuration.check('outputDirectory')):
        v = app.exec_()

    vistrails.gui.application.stop_application()
    sys.exit(v)

if __name__ == '__main__':
    main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.modules.vistrails_module import ModuleError
from fold import Fold, FoldWithModule

#################################################################################
## Some useful loop structures

class Map(FoldWithModule):
    """A Map module, that just appends the results in a list."""

    def setInitialValue(self):
        """Defining the initial value..."""

        self.initialValue = []

    def operation(self):
        """Defining the operation..."""

        self.partialResult.append(self.elementResult)


class Filter(FoldWithModule):
    """A Filter module, that returns in a list only the results that satisfy a
    condition."""

    def setInitialValue(self):
        """Defining the initial value..."""

        self.initialValue = []

    def operation(self):
        """Defining the operation..."""

        if not isinstance(self.elementResult, bool):
            raise ModuleError(self,'The function applied to the elements of the\
list must return a boolean result.')

        if self.elementResult:
            self.partialResult.append(self.element)


class Sum(Fold):
    """A Sum module, that computes the sum of the elements in a list."""

    def setInitialValue(self):
        """Defining the initial value..."""

        self.initialValue = 0

    def operation(self):
        """Defining the operation..."""

        self.partialResult += self.element


class And(Fold):
    """An And module, that computes the And result among the elements
    in a list."""

    def setInitialValue(self):
        """Defining the initial value..."""

        self.initialValue = True

    def operation(self):
        """Defining the operation..."""

        self.partialResult = self.partialResult and bool(self.element)


class Or(Fold):
    """An Or module, that computes the Or result among the elements
    in a list."""

    def setInitialValue(self):
        """Defining the initial value..."""

        self.initialValue = False

    def operation(self):
        """Defining the operation..."""

        self.partialResult = self.partialResult or self.element


###############################################################################

import unittest
import urllib2

from vistrails.tests.utils import intercept_result, execute


class TestMap(unittest.TestCase):
    def test_simple(self):
        src = urllib2.quote('o = i + 1')
        with intercept_result(Map, 'Result') as results:
            self.assertFalse(execute([
                    ('PythonSource', 'org.vistrails.vistrails.basic', [
                        ('source', [('String', src)]),
                    ]),
                    ('Map', 'org.vistrails.vistrails.control_flow', [
                        ('InputPort', [('List', "['i']")]),
                        ('OutputPort', [('String', 'o')]),
                        ('InputList', [('List', '[1, 2, 8, 9.1]')]),
                    ]),
                ],
                [
                    (0, 'self', 1, 'FunctionPort'),
                ],
                add_port_specs=[
                    (0, 'input', 'i',
                     'org.vistrails.vistrails.basic:Float'),
                    (0, 'output', 'o',
                     'org.vistrails.vistrails.basic:Float'),
                ]))
        self.assertEqual(results, [[2, 3, 9, 10.1]])

    def test_tuple(self):
        src = urllib2.quote('o = len(i[0]) + i[1]')
        with intercept_result(Map, 'Result') as results:
            self.assertFalse(execute([
                    ('PythonSource', 'org.vistrails.vistrails.basic', [
                        ('source', [('String', src)]),
                    ]),
                    ('Map', 'org.vistrails.vistrails.control_flow', [
                        ('InputPort', [('List', "['i']")]),
                        ('OutputPort', [('String', 'o')]),
                        ('InputList', [('List',
                            '[("aa", 1), ("", 8), ("a", 4)]')]),
                    ]),
                ],
                [
                    (0, 'self', 1, 'FunctionPort'),
                ],
                add_port_specs=[
                    (0, 'input', 'i',
                     'org.vistrails.vistrails.basic:String,org.vistrails.vistrails.basic:Integer'),
                    (0, 'output', 'o',
                     'org.vistrails.vistrails.basic:Float'),
                ]))
        self.assertEqual(results, [[3, 8, 5]])

    def test_multiple(self):
        src = urllib2.quote('o = i + j')
        with intercept_result(Map, 'Result') as results:
            self.assertFalse(execute([
                    ('PythonSource', 'org.vistrails.vistrails.basic', [
                        ('source', [('String', src)]),
                    ]),
                    ('Map', 'org.vistrails.vistrails.control_flow', [
                        ('InputPort', [('List', "['i', 'j']")]),
                        ('OutputPort', [('String', 'o')]),
                        ('InputList', [('List',
                            '[(1, 2), (3, 8), (-2, 3)]')]),
                    ]),
                ],
                [
                    (0, 'self', 1, 'FunctionPort'),
                ],
                add_port_specs=[
                    (0, 'input', 'i',
                     'org.vistrails.vistrails.basic:Integer'),
                    (0, 'input', 'j',
                     'org.vistrails.vistrails.basic:Integer'),
                    (0, 'output', 'o',
                     'org.vistrails.vistrails.basic:Integer'),
                ]))
        self.assertEqual(results, [[3, 11, 1]])


class TestUtils(unittest.TestCase):
    def test_filter(self):
        src = urllib2.quote('o = bool(i)')
        with intercept_result(Filter, 'Result') as results:
            self.assertFalse(execute([
                    ('PythonSource', 'org.vistrails.vistrails.basic', [
                        ('source', [('String', src)]),
                    ]),
                    ('Filter', 'org.vistrails.vistrails.control_flow', [
                        ('InputPort', [('List', "['i']")]),
                        ('OutputPort', [('String', 'o')]),
                        ('InputList', [('List',
                            "[0, 1, 2, 3, '', 'foo', True, False]")]),
                    ]),
                ],
                [
                    (0, 'self', 1, 'FunctionPort'),
                ],
                add_port_specs=[
                    (0, 'input', 'i',
                     'org.vistrails.vistrails.basic:Module'),
                    (0, 'output', 'o',
                     'org.vistrails.vistrails.basic:Boolean'),
                ]))
        self.assertEqual(results, [[1, 2, 3, 'foo', True]])

    def test_sum(self):
        with intercept_result(Sum, 'Result') as results:
            self.assertFalse(execute([
                    ('Sum', 'org.vistrails.vistrails.control_flow', [
                        ('InputList', [('List', "[1, 2, 3, 8, 14.7]")]),
                    ]),
                ]))
        self.assertEqual(results, [28.7])

    def do_andor(self, l):
        with intercept_result(And, 'Result') as and_results:
            with intercept_result(Or, 'Result') as or_results:
                self.assertFalse(execute([
                        ('List', 'org.vistrails.vistrails.basic', [
                            ('value', [('List', str(l))]),
                        ]),
                        ('And', 'org.vistrails.vistrails.control_flow', []),
                        ('Or', 'org.vistrails.vistrails.control_flow', []),
                    ],
                    [
                        (0, 'value', 1, 'InputList'),
                        (0, 'value', 2, 'InputList'),
                    ]))
        self.assertEqual(len(and_results), 1)
        self.assertEqual(len(or_results), 1)
        return and_results[0], or_results[0]

    def test_andor(self):
        self.assertEqual(self.do_andor([False, False]), (False, False))
        self.assertEqual(self.do_andor([True, False]), (False, True))
        self.assertEqual(self.do_andor([True, True]), (True, True))
        self.assertEqual(self.do_andor([False, True]), (False, True))

        # This is kind of random
        self.assertEqual(self.do_andor([]), (True, False))



#!/usr/bin/env python

# This simple example shows how to do basic rendering and pipeline
# creation. It also demonstrates the use of the LODActor.

# Import the VTK-Python module
import vtk
from vtk.util.misc import vtkGetDataRoot
from vtk.util.colors import *
# Get the location of the data.
VTK_DATA_ROOT = vtkGetDataRoot()

# This creates a polygonal cylinder model with eight circumferential
# facets.
part = vtk.vtkSTLReader()
part.SetFileName(VTK_DATA_ROOT + "/Data/42400-IDGH.stl")

# The mapper is responsible for pushing the geometry into the graphics
# library. It may also do color mapping, if scalars or other
# attributes are defined.
partMapper = vtk.vtkPolyDataMapper()
partMapper.SetInputConnection(part.GetOutputPort())

# The LOD actor is a special type of actor. It will change appearance
# in order to render faster. At the highest resolution, it renders
# ewverything just like an actor. The middle level is a point cloud,
# and the lowest level is a simple bounding box.
partActor = vtk.vtkLODActor()
partActor.SetMapper(partMapper)
partActor.GetProperty().SetColor(light_grey)
partActor.RotateX(30.0)
partActor.RotateY(-45.0)

# Create the graphics structure. The renderer renders into the render
# window. The render window interactor captures mouse events and will
# perform appropriate camera or actor manipulation depending on the
# nature of the events.
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(partActor)
ren.SetBackground(0.1, 0.2, 0.4)
renWin.SetSize(200, 200)

# We'll zoom in a little by accessing the camera and invoking a "Zoom"
# method on it.
ren.ResetCamera()
ren.GetActiveCamera().Zoom(1.5)

# This starts the event loop.
iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import urllib
from vistrails.core.mashup import conv_from_bool, conv_to_bool, convert_symbols
from vistrails.db.domain import DBMashupComponent

import unittest
from vistrails.db.domain import IdScope
import copy

################################################################################
class Component(DBMashupComponent):
    def __init__(self, id, vttype, param_id, parent_vttype, parent_id, mid, 
                 type, value, p_pos, pos, strvaluelist, minVal="0",
                 maxVal="1", stepSize="1", parent='', seq=0, widget="text"):
    
        """Component() 
        widget can be: text, slider, combobox, numericstepper, checkbox

        """
        DBMashupComponent.__init__(self, id, param_id, vttype, parent_vttype, 
                                   parent_id, p_pos, mid, pos, type, value, 
                                   minVal, maxVal, stepSize, strvaluelist, 
                                   widget, seq, parent)
        if isinstance(seq, bool):
            self.seq = seq

    id = DBMashupComponent.db_id
    vttype = DBMashupComponent.db_vttype
    vtid = DBMashupComponent.db_vtid
    vtparent_type = DBMashupComponent.db_vtparent_type
    vtparent_id = DBMashupComponent.db_vtparent_id
    vtmid = DBMashupComponent.db_vtmid
    vtpos = DBMashupComponent.db_vtpos
    pos = DBMashupComponent.db_pos
    type = DBMashupComponent.db_type
    minVal = DBMashupComponent.db_minVal
    maxVal = DBMashupComponent.db_maxVal
    stepSize = DBMashupComponent.db_stepSize
    parent = DBMashupComponent.db_parent
    widget = DBMashupComponent.db_widget
    strvaluelist = DBMashupComponent.db_strvaluelist
    
    def _get_seq(self):
        return conv_to_bool(self.db_seq)
    def _set_seq(self, s):
        self.db_seq = conv_from_bool(s)
    seq = property(_get_seq,_set_seq)
     
    def _get_val(self):
        self.db_val = convert_symbols(self.db_val)
        return self.db_val
    def _set_val(self, v):
        self.db_val = v
    val = property(_get_val,_set_val)
    
    def _get_valuelist(self):
        self.strvaluelist = convert_symbols(self.strvaluelist)
        data = self.strvaluelist.split(',')
        result = []
        for d in data:
            result.append(urllib.unquote_plus(d))
        return result
    def _set_valuelist(self, valuelist):
        q = []
        for v in valuelist:
            q.append(urllib.quote_plus(v))
        self.strvaluelist = ",".join(q)
    valueList = property(_get_valuelist,_set_valuelist)

    @staticmethod
    def convert(_component):
        _component.__class__ = Component
        
    def __copy__(self):
        return Component.do_copy(self)
    
    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        """do_copy() -> Component 
        returns a clone of itself"""
        cp = DBMashupComponent.do_copy(self, new_ids, id_scope, id_remap)      
        Component.convert(cp)
        return cp
    
    ##########################################################################
    # Serialization / Unserialization
    
#    def toXml(self, node=None):
#        """toXml(node: ElementTree.Element) -> ElementTree.Element
#             writes itself to xml
#        """
#        if node is None:
#            node = ElementTree.Element('component')
#        #set attributes
#        node.set('id', self.convert_to_str(self.id,'long'))
#        node.set('vttype', self.convert_to_str(self.vttype,'str'))
#        node.set('vtid', self.convert_to_str(self.vtid,'long'))
#        node.set('vtparent_type', self.convert_to_str(self.vtparent_type,'str'))
#        node.set('vtparent_id', self.convert_to_str(self.vtparent_id,'long'))
#        node.set('vtmid', self.convert_to_str(self.vtmid,'long'))
#        node.set('vtpos', self.convert_to_str(self.vtpos,'long'))
#        
#        node.set('pos', self.convert_to_str(self.pos,'long'))
#        node.set('type', self.convert_to_str(self.type,'str'))
#        
#        node.set('val', self.convert_to_str(self.val, 'str'))
#        node.set('minVal', self.convert_to_str(self.minVal,'str'))
#        node.set('maxVal', self.convert_to_str(self.maxVal,'str'))
#        node.set('stepSize', self.convert_to_str(self.stepSize,'str'))
#        node.set('valueList',self.convert_to_str(self.strvaluelist,'str'))
#        node.set('parent', self.convert_to_str(self.parent,'str'))
#        node.set('seq', self.convert_to_str(self.seq,'bool'))
#        node.set('widget',self.convert_to_str(self.widget,'str'))
#        return node
#
#    @staticmethod
#    def fromXml(node):
#        if node.tag != 'component':
#            return None
#
#        #read attributes
#        data = node.get('id', None)
#        id = Component.convert_from_str(data, 'long')
#        data = node.get('vttype', None)
#        vttype = Component.convert_from_str(data, 'str')
#        data = node.get('vtid', None)
#        vtid = Component.convert_from_str(data, 'long')
#        data = node.get('vtparent_type', None)
#        vtparent_type = Component.convert_from_str(data, 'str')
#        data = node.get('vtparent_id', None)
#        vtparent_id = Component.convert_from_str(data, 'long')
#        data = node.get('vtmid', None)
#        vtmid = Component.convert_from_str(data, 'long')
#        data = node.get('vtpos', None)
#        vtpos = Component.convert_from_str(data, 'long')
#        data = node.get('pos', None)
#        pos = Component.convert_from_str(data, 'long')
#        data = node.get('type', None)
#        type = Component.convert_from_str(data, 'str')
#        data = node.get('val', None)
#        val = Component.convert_from_str(data, 'str')
#        val = val.replace("&lt;", "<")
#        val = val.replace("&gt;", ">")
#        val = val.replace("&amp;","&")
#        data = node.get('minVal', None)
#        minVal = Component.convert_from_str(data, 'str')
#        data = node.get('maxVal', None)
#        maxVal = Component.convert_from_str(data, 'str')
#        data = node.get('stepSize', None)
#        stepSize = Component.convert_from_str(data, 'str')
#        data = node.get('valueList', None)
#        values = Component.convert_from_str(data, 'str')
#        values = values.replace("&lt;", "<")
#        values = values.replace("&gt;", ">")
#        values = values.replace("&amp;","&")
#        data = node.get('parent', None)
#        parent = Component.convert_from_str(data, 'str')
#        data = node.get('seq', None)
#        seq = Component.convert_from_str(data, 'bool')
#        data = node.get('widget', None)
#        widget = Component.convert_from_str(data, 'str')
#       
#        component = Component(id=id, vttype=vttype, param_id=vtid, 
#                              parent_vttype=vtparent_type, 
#                              parent_id=vtparent_id, mid=vtmid, type=type,
#                              value=val, p_pos=vtpos, pos=pos,
#                              minVal=minVal,
#                              maxVal=maxVal,
#                              stepSize=stepSize,
#                              strvaluelist=values,
#                              parent=parent,
#                              seq=seq,
#                              widget=widget)
#        return component

    ##########################################################################
    # Operators

    def __str__(self):
        """ __str__() -> str - Returns a string representation of itself """
        
        return ("(Component id='%s' vttype='%s' vtid='%s' vtparent_type='%s' \
vtparent_id='%s' vtmid='%s' vtpos='%s' type='%s' pos='%s' val='%s' minVal='%s' \
maxVal='%s' stepSize='%s' strvaluelist='%s' parent='%s' seq='%s' widget='%s')@%X" %
                    (self.id,
                     self.vttype,
                     self.vtid,
                     self.vtparent_type,
                     self.vtparent_id,
                     self.vtmid,
                     self.vtpos,
                     self.type,
                     self.pos,
                     self.val,
                     self.minVal,
                     self.maxVal,
                     self.stepSize,
                     self.strvaluelist,
                     self.parent,
                     self.seq,
                     self.widget,
                     id(self)))

    def __eq__(self, other):
        """ __eq__(other: Component) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.vttype != other.vttype:
            return False
        if self.vtid != other.vtid:
            return False
        if self.vtparent_type != other.vtparent_type:
            return False
        if self.vtparent_id != other.vtparent_id:
            return False
        if self.vtmid != other.vtmid:
            return False
        if self.vtpos != other.vtpos:
            return False
        if self.type != other.type:
            return False
        if self.pos != other.pos:
            return False
        if self.val != other.val:
            return False
        if self.minVal != other.minVal:
            return False
        if self.maxVal != other.maxVal:
            return False
        if self.stepSize != other.stepSize:
            return False
        if self.strvaluelist != other.strvaluelist:
            return False
        if self.parent != other.parent:
            return False
        if self.seq != other.seq:
            return False
        if self.widget != other.widget:
            return False
        return True

    def __ne__(self, other):
        """ __ne__(other: Component) -> boolean
        Returns True if self and other don't have the same attributes. 
        Used by !=  operator. 
        
        """
        return not self.__eq__(other)
    
################################################################################


class TestComponent(unittest.TestCase):
    def create_component(self, id_scope=IdScope()):
        c = Component(id=id_scope.getNewId('mashup_component'),
                          vttype='parameter', param_id=15L, 
                          parent_vttype='function', parent_id=3L, mid=4L,
                          type='String', value='test', p_pos=0, pos=1, 
                          strvaluelist='test1,test2', widget="text")
        return c
    
    def test_copy(self):
        id_scope = IdScope()
        c1 = self.create_component(id_scope)
        c2 = copy.copy(c1)
        self.assertEqual(c1, c2)
        self.assertEqual(c1.id, c2.id)
        c3 = c2.do_copy(True, id_scope, {})
        self.assertEqual(c1,c3)
        self.assertNotEqual(c1.id, c3.id)
        
    def test_valuelist(self):
        c1 = self.create_component()
        c1.strvaluelist = "1,2,3"
        self.assertEqual(['1','2','3'], c1.valueList)
        c1.valueList = ['1','2','3']
        self.assertEqual(c1.strvaluelist,"1,2,3")
        
#        #testing values with , after serialization
#        c1.valueList = ['a,b,c', '123', ',as']
#        c2 = Component.fromXml(c1.toXml())
#        self.assertEqual(c1.strvaluelist, c2.strvaluelist)
#        self.assertEqual(c1.valueList, c2.valueList)
#        self.assertEqual(c2.valueList, ['a,b,c', '123', ',as'])
        
    def test_str(self):
        c1 = self.create_component()
        str(c1)
        